import subprocess
import sys
from typing import Any, Dict, List, Optional, Tuple

import click
from rich import box, table

from coalescenceml.config.profile_config import ProfileConfiguration
from coalescenceml.constants import console
from coalescenceml.directory import Directory
from coalescenceml.enums import StackComponentFlavor
from coalescenceml.integrations.integration import IntegrationMeta
from coalescenceml.logger import get_logger
from coalescenceml.stack import StackComponent


logger = get_logger(__name__)


def title(text: str) -> None:
    """Echo a title formatted string on the CLI.

    Args:
        text: Input text string
    """
    console.print(text.upper(), style="title")
    # click.echo(click.style(text.upper(), fg="cyan", bold=True, underline=True))


def confirmation(text: str, *args: Any, **kwargs: Any) -> bool:
    """Echo a confirmation string on the CLI.

    Args:
        text: Input text string.
        *args: Args to be passed to click.confirm()
        **kwargs: Kwargs to be passed to click.confirm()

    Returns:
        Boolean based on user response
    """
    return click.confirm(click.style(text, fg="yellow"), *args, **kwargs)


def error(text: str) -> None:
    """Echo an error string on the CLI.

    Args:
        text: Input text string

    Raises:
        ClickException: when called.
    """
    raise click.ClickException(message=click.style(text, fg="red", bold=True))


def warning(text: str) -> None:
    """Echo a warning string on the CLI.

    Args:
        text: Input text string.
    """
    click.echo(click.style(text, fg="yellow", bold=True))


def info(text: str) -> None:
    """Echo an info dump on the CLI.

    Args:
        text: Input text string.
    """
    console.print(text, style="info")
    # click.echo(click.style(text, fg="green"))


def print_table(tab: List[Dict[str, Any]]) -> None:
    """Prints the list of dicts in a table format.

    ..note: Each item in the list represents a line in the table. Each dict
    should have the same keys.

    ..warning: Undefined behavior if list is empty!

    Args:
        tab: A table represented as a list of dicts
    """
    columns = {key.upper(): None for key in tab[0].keys()}
    rich_table = table.Table(*columns.keys(), box=box.HEAVY_EDGE)

    for dict_ in tab:
        values = columns.copy()
        values.update(dict_)
        rich_table.add_row(*list(values.values()))

    if len(rich_table.columns) > 1:
        rich_table.columns[0].justify = "center"
    console.print(rich_table)


def format_integration_list(
    integrations: List[Tuple[str, IntegrationMeta]]
) -> List[Dict[str, str]]:
    """Formats a list of integrations into a List of Dicts."""

    list_of_dicts = []
    for name, integration_impl in integrations:
        is_installed = integration_impl.check_installation()  # type: ignore[attr-defined]
        list_of_dicts.append(
            {
                "INSTALLED": ":white_check_mark:" if is_installed else "",
                "INTEGRATION": name,
                "REQUIRED_PACKAGES": ", ".join(integration_impl.REQUIREMENTS),  # type: ignore[attr-defined]
            }
        )
    return list_of_dicts


def print_stack_component_list(
    components: List[StackComponent],
    active_component_name: Optional[str] = None,
) -> None:
    """Prints a table with configuration options for a list of stack components.

    ..note: If a component is active (its name matches the
    `active_component_name`), it will be highlighted in a separate table column.

    Args:
        components: List of stack components to print.
        active_component_name: Name of the component that is currently
            active.
    """
    configurations = []
    for component in components:
        is_active = component.name == active_component_name
        component_config = {
            "ACTIVE": ":point_right:" if is_active else "",
            **{
                key.upper(): str(value)
                for key, value in component.dict().items()
            },
        }
        configurations.append(component_config)
    print_table(configurations)


def print_stack_configuration(
    config: Dict[StackComponentFlavor, str], active: bool, stack_name: str
) -> None:
    """Prints the configuration options of a stack.

    Args:
        config:
        active:
        stack_name:
    """
    stack_caption = f"'{stack_name}' stack"
    if active:
        stack_caption += " (ACTIVE)"
    rich_table = table.Table(
        box=box.HEAVY_EDGE,
        title="Stack Configuration",
        caption=stack_caption,
        show_lines=True,
    )
    rich_table.add_column("COMPONENT_TYPE")
    rich_table.add_column("COMPONENT_NAME")
    for component_type, name in config.items():
        rich_table.add_row(component_type.value, name)

    # capitalize entries in first column
    rich_table.columns[0]._cells = [
        component.upper() for component in rich_table.columns[0]._cells
    ]
    console.print(rich_table)


def print_stack_component_configuration(
    component: StackComponent, display_name: str, active_status: bool
) -> None:
    """Prints the configuration options of a stack component."""
    title = f"{component.TYPE.value.upper()} Component Configuration"
    if active_status:
        title += " (ACTIVE)"
    rich_table = table.Table(
        box=box.HEAVY_EDGE,
        title=title,
        show_lines=True,
    )
    rich_table.add_column("COMPONENT_PROPERTY")
    rich_table.add_column("VALUE")
    items = component.dict().items()
    for item in items:
        rich_table.add_row(*[str(elem) for elem in item])

    # capitalize entries in first column
    rich_table.columns[0]._cells = [
        component.upper() for component in rich_table.columns[0]._cells
    ]
    console.print(rich_table)


def print_active_profile() -> None:
    """Print active profile."""
    dir_ = Directory()
    scope = "local" if dir_.root else "global"
    info(f"Running with active profile: '{dir_.active_profile_name}' ({scope})")


def print_active_stack() -> None:
    """Print active stack."""
    dir_ = Directory()
    info(f"Running with active stack: '{dir_.active_stack_name}'")


def print_profile(
    profile: ProfileConfiguration,
    active: bool,
) -> None:
    """Prints the configuration options of a profile.
    Args:
        profile: Profile to print.
        active: Whether the profile is active.
        name: Name of the profile.
    """
    profile_title = f"'{profile.name}' Profile Configuration"
    if active:
        profile_title += " (ACTIVE)"

    rich_table = table.Table(
        box=box.HEAVY_EDGE,
        title=profile_title,
        show_lines=True,
    )
    rich_table.add_column("PROPERTY")
    rich_table.add_column("VALUE")
    items = profile.dict().items()
    for item in items:
        rich_table.add_row(*[str(elem) for elem in item])

    # capitalize entries in first column
    rich_table.columns[0]._cells = [
        component.upper() for component in rich_table.columns[0]._cells
    ]
    console.print(rich_table)


def parse_unknown_options(args: List[str]) -> Dict[str, Any]:
    """Parse unknown options from the CLI.
    Args:
      args: A list of strings from the CLI.
    Returns:
        Dict of parsed args.
    """
    warning_message = (
        "Please provide args with a proper "
        "identifier as the key and the following structure: "
        '--custom_argument="value"'
    )

    assert all(a.startswith("--") for a in args), warning_message
    assert all(len(a.split("=")) == 2 for a in args), warning_message

    p_args = [a.lstrip("--").split("=") for a in args]

    assert all(k.isidentifier() for k, _ in p_args), warning_message

    r_args = {k: v for k, v in p_args}
    assert len(p_args) == len(r_args), "Replicated arguments!"

    return r_args


def install_packages(packages: List[str]) -> None:
    """Installs pypi package into the current environment with pip.

    Args:
        package: pypi package to install
    """
    # TODO: Can we utilize a package manager instead? Such as their package
    # manager so that we don't create conflicts?
    # IF we enforce the usage of `poetry` then this is do-able.
    # subprocess.check_call(["poetry", "add", package])
    subprocess.check_call([sys.executable, "-m", "pip", "install"] + packages)


def uninstall_package(package: str) -> None:
    """Uninstalls pypi package from the current environment with pip.

    Args:
        package: pypi package to install
    """
    # TODO: Can we utilize a package manager instead? Such as their package
    # manager so that we don't create conflicts?
    # IF we enforce the usage of `poetry` then this is do-able.
    # subprocess.check_call(["poetry", "remove", package])
    subprocess.check_call(
        [sys.executable, "-m", "pip", "uninstall", "-y", package]
    )



from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict

from coalescenceml.enums import StackComponentFlavor
from coalescenceml.stack import StackComponent


class BaseExperimentTracker(StackComponent, ABC):
    """Base class for all CoalescenceML experiment trackers."""

    # Class config
    TYPE: ClassVar[StackComponentFlavor] = StackComponentFlavor.EXPERIMENT_TRACKER
    FLAVOR: ClassVar[str]


    @abstractmethod
    def log_params(self, params: Dict[str, Any]) -> None:
        raise NotImplementedError()

    @abstractmethod
    def log_metrics(self, metrics: Dict[str, Any]) -> None:
        raise NotImplementedError()

    @abstractmethod
    def log_artifacts(self, artifacts: Dict[str, Any]) -> None:
        raise NotImplementedError()



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, NamedTuple, Optional, Type, cast

from coalescenceml.directory import Directory
from coalescenceml.step.exceptions import StepContextError


if TYPE_CHECKING:
    from coalescenceml.artifacts.base_artifact import BaseArtifact
    from coalescenceml.metadata_store.base_metadata_store import (
        BaseMetadataStore,
    )
    from coalescenceml.producers.base_producer import BaseProducer
    from coalescenceml.stack import Stack


class StepContextOutput(NamedTuple):
    """"""

    producer_class: Type[BaseProducer]
    artifact: BaseArtifact


class StepContext(object):
    """"""

    def __init__(
        self,
        step_name: str,
        output_producers: Dict[str, Type[BaseProducer]],
        output_artifacts: Dict[str, BaseArtifact],
    ):
        """ """
        if output_producers.keys() != output_artifacts.keys():
            raise ValueError(f"Mismatched keys")

        self.step_name = step_name
        self._outputs = {
            key: StepContextOutput(
                output_producer[key],
                output_artifact[key],
            )
            for key in output_artifacts.keys()
        }
        self._metadata_store = Directory().active_stack.metadata_store
        self._stack = Directory().active_stack

    def _get_output(
        self, output_name: Optional[str] = None
    ) -> StepContextOutput:
        """Returns the producer and artifact URI for a given step output.

        Args:
            output_name: Optional name of the output for which to get the
                producer and URI.

        Returns:
            Tuple containing the producer and artifact URI for the
            given output.

        Raises:
            StepContextError: If the step has no outputs, no output for
                              the given `output_name` or if no `output_name`
                              was given but the step has multiple outputs.
        """
        output_count = len(self._outputs)
        if output_count == 0:
            raise StepContextError(
                f"Unable to get step output for step '{self.step_name}': "
                f"This step does not have any outputs."
            )

        if not output_name and output_count > 1:
            raise StepContextError(
                f"Unable to get step output for step '{self.step_name}': "
                f"This step has multiple outputs ({set(self._outputs)}), "
                f"please specify which output to return."
            )

        if output_name:
            if output_name not in self._outputs:
                raise StepContextError(
                    f"Unable to get step output '{output_name}' for "
                    f"step '{self.step_name}'. This step does not have an "
                    f"output with the given name, please specify one of the "
                    f"available outputs: {set(self._outputs)}."
                )
            return self._outputs[output_name]
        else:
            return next(iter(self._outputs.values()))

    @property
    def metadata_store(self) -> BaseMetadataStore:
        """
        Returns an instance of the metadata store that is used to store
        metadata about the step (and the corresponding pipeline) which is
        being executed.
        """
        return self._metadata_store

    @property
    def stack(self) -> Optional["Stack"]:
        """Returns the current active stack."""
        return self._stack

    def get_output_producer(
        self,
        output_name: Optional[str] = None,
        custom_producer_class: Optional[Type[Baseproducer]] = None,
    ) -> Baseproducer:
        """Returns a producer for a given step output.

        Args:
            output_name: Optional name of the output for which to get the
                producer. If no name is given and the step only has a
                single output, the producer of this output will be
                returned. If the step has multiple outputs, an exception
                will be raised.
            custom_producer_class: If given, this `Baseproducer`
                subclass will be initialized with the output artifact instead
                of the producer that was registered for this step output.

        Returns:
            A producer initialized with the output artifact for
            the given output.

        Raises:
            StepContextError: If the step has no outputs, no output for
                              the given `output_name` or if no `output_name`
                              was given but the step has multiple outputs.
        """
        producer_class, artifact = self._get_output(output_name)
        # use custom producer class if provided or fallback to default
        # producer for output
        producer_class = custom_producer_class or producer_class
        return producer_class(artifact)

    def get_output_artifact_uri(self, output_name: Optional[str] = None) -> str:
        """Returns the artifact URI for a given step output.

        Args:
            output_name: Optional name of the output for which to get the URI.
                If no name is given and the step only has a single output,
                the URI of this output will be returned. If the step has
                multiple outputs, an exception will be raised.

        Returns:
            Artifact URI for the given output.

        Raises:
            StepContextError: If the step has no outputs, no output for
                              the given `output_name` or if no `output_name`
                              was given but the step has multiple outputs.
        """
        return cast(str, self._get_output(output_name).artifact.uri)



from coalescenceml.integrations.constants import AWS
from coalescenceml.integrations.integration import Integration


class AWSIntegration(Integration):
    """Integration for AWS."""

    NAME = AWS
    REQUIREMENTS = ["s3fs"]

    @classmethod
    def activate(cls) -> None:
        from coalescenceml.integrations.aws_s3 import artifact_store


AWSIntegration.check_installation()


import os
from typing import Any, ClassVar, Dict, Optional

import mlflow
from mlflow.entities import Experiment
from pydantic import root_validator, validator


from coalescenceml.directory import Directory
from coalescenceml.environment import Environment
from coalescenceml.experiment_tracker import (
    BaseExperimentTracker,
)
from coalescenceml.integrations.constants import MLFLOW
from coalescenceml.logger import get_logger
from coalescenceml.stack.stack_component_class_registry import (
    register_stack_component_class,
)
from coalescenceml.stack.stack_validator import StackValidator

logger = get_logger(__name__)

MLFLOW_TRACKING_TOKEN = "MLFLOW_TRACKING_TOKEN"
MLFLOW_TRACKING_USERNAME = "MLFLOW_TRACKING_USERNAME"
MLFLOW_TRACKING_PASSWORD = "MLFLOW_TRACKING_PASSWORD"
# TODO: Determine whether to use insecure TLS
# TODO: How do we use a cert bundle if desired?
# https://www.mlflow.org/docs/latest/tracking.html#logging-to-a-tracking-server


@register_stack_component_class
class MLFlowExperimentTracker(BaseExperimentTracker):
    """Stores MLFlow Configuration and interaction functions.

    CoalescenceML configures MLFlow for you....
    """
    # TODO: Ask Iris about local mlflow runs
    tracking_uri: Optional[str] = None # Can this
    use_local_backend: bool = False # base it on tracking_uri
    tracking_token: Optional[str] = None # The way I prefer using MLFlow
    tracking_username: Optional[str] = None
    tracking_password: Optional[str] = None

    FLAVOR: ClassVar[str] = MLFLOW

    @validator("tracking_uri")
    def ensure_valid_tracking_uri(
        cls, tracking_uri: Optional[str] = None,
    ) -> Optional[str]:
        """Ensure the tracking uri is a valid mlflow tracking uri.

        Args:
            tracking_uri: The value to verify

        Returns:
            Valid tracking uri

        Raises:
            ValueError: if mlflow tracking uri is invalid.
        """
        if tracking_uri:
            valid_protocols = DATABASE_ENGINES + ["http", "https", "file"]
            if not any(
                tracking_uri.startswith(protocol)
                for protocol in valid_protocols
            ):
                raise ValueError(
                    f"MLFlow tracking uri does not use a valid protocol "
                    f" which inclue: {valid_protocols}. Please see "
                    f"https://www.mlflow.org/docs/latest/tracking.html"
                    f"#where-runs-are-recorded for more information."
                )

        return tracking_uri

    @staticmethod
    def is_remote_tracking_uri(tracking_uri: str) -> bool:
        """Check whether URI is using remote protocol.

        Args:
            tracking_uri: MLFlow tracking server location

        Returns:
            True if server is remote else False
        """
        return any(
            tracking_uri.startswith(prefix)
            for prefix in ["http", "https"]
            # Only need these as tested with AWS MySQL generates hostname
            # with HTTP. Hence any cloud servers will have to have http and 
            # local ones do not. But this might be wrong and need to change.
        )

    @root_validator
    def ensure_tracking_uri_or_local(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Ensure that the tracking uri exists or use local backend is true.

        Args:
            values: Pydantic stored values of class

        Returns:
            Pydantic class values
        """
        tracking_uri = values.get("tracking_uri")
        use_local_backend = values.get("use_local_backend")

        if not (tracking_uri or use_local_backend):
            raise ValueError(
                f"You must specify either a tracking uri or the use "
                f"your local artifact store for the MLFlow "
                f"experiment tracker."
                # TODO: Message on how to fix.
            )

        return values


    @root_validator
    def ensure_authentication(
        cls, values: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Ensure that credentials exists for a remote tracking instance.

        Args:
            values: Pydantic stored values of class

        Returns:
            class pydantic values
        """
        tracking_uri = values.get("tracking_uri")

        if tracking_uri: # Check if exists
            if cls.is_remote_tracking_uri(tracking_uri): # Check if remote
                token_auth = values.get("tracking_token")

                basic_http_auth = (
                    values.get("tracking_username")
                    and values.get("tracking_password")
                )

                if not (token_auth or basic_http_auth):
                    raise ValueError(
                        f"MLFlow experiment tracking with a remote tracking "
                        f"uri '{tracking_uri}' is only allowed when either a "
                        f"username and password or auth token is used."
                        # TODO: Add update commands to stack to all users to update the component.
                    )

        return values

    @staticmethod
    def local_mlflow_tracker() -> str:
        """Return local mlflow folder inside artifact store.

        Returns:
            MLFlow tracking uri for local backend.
        """
        dir_ = Directory(skip_directory_check=True)
        artifact_store = dir_.active_stack.artifact_store
        # TODO: MLFlow can connect to non-local stores however
        # I am unsure what this entails and how to test this.
        # AWS would be an interesting testbed for a future iteration.
        # Ideally this would work for an arbitrary artifact store.

        local_mlflow_uri = os.path.join(artifact_store.path, "mlruns")
        if not os.path.exists(local_mlflow_uri):
            os.makedirs(local_mlflow_uri)

        return "file:" + local_mlflow_uri

    def get_tracking_uri(self) -> str:
        """Return configured MLFlow tracking uri.

        Returns:
            MLFlow tracking uri
        
        Raises:
            ValueError: if tracking uri is empty and user didn't specify to use the local backend.
        """
        if self.tracking_uri:
            return self.tracking_uri
        else:
            if self.use_local_backend:
                return self.local_mlflow_tracker()
            else:
                raise ValueError(
                    f"You must specify either a tracking uri or the use "
                    f"of your local artifact store for the MLFlow "
                    f"experiment tracker."
                )

    def prepare_step_run(self) -> None:
        """Configure MLFlow tracking URI and credentials."""
        if self.tracking_token:
            os.environ[MLFLOW_TRACKING_TOKEN] = self.tracking_token
        if self.tracking_username:
            os.environ[MLFLOW_TRACKING_USERNAME] = self.tracking_username
        if self.tracking_password:
            os.environ[MLFLOW_TRACKING_PASSWORD] = self.tracking_password

        mlflow.set_tracking_uri(self.get_tracking_uri())
        return

    def cleanup_step_run(self) -> None:
        mlflow.set_tracking_uri("")

    @property
    def validator(self) -> Optional[StackValidator]:
        """Validate that MLFlow config with the rest of stack is valid.

        ..note: We just need to check (for now) that if the use local flag
            is used that there is a local artifact store
        """
        if self.tracking_uri:
            # Tracking URI exists so do nothing
            return None
        else:
            from coalescenceml.artifact_store import LocalArtifactStore
            # Presumably they've set the use_local_backend to true
            # So check for local artifact store b/c thats all that
            # works for now. This will be edited later (probably...)
            return StackValidator(
                custom_validation_function=lambda stack: (
                    isinstance(stack.artifact_store, LocalArtifactStore),
                    "MLFlow experiment tracker with local backend only "
                    "works with a local artifact store at this time."
                )
            )


    def active_experiment(self, experiment_name=None) -> Optional[Experiment]:
        """Return currently active MLFlow experiment.

        Args:
            experiment_name: experiment name to set in MLFlow.

        Returns:
            None if not in step else will return an MLFlow Experiment
        """
        step_env = Environment().step_env

        if not step_env:
            # I.e. we are not in a step running
            return None

        experiment_name = experiment_name or step_env.pipeline_name
        mlflow.set_experiment(experiment_name=experiment_name)
        return mlflow.get_experiment_by_name(experiment_name)

    def active_run(self, experiment_name=None) -> Optional[mlflow.ActiveRun]:
        """"""
        step_env = Environment().step_env
        active_experiment = self.active_experiment(experiment_name)

        if not active_experiment:
            # This checks for experiment + not being in setp
            # Should we make this explicit or keep it as implicit?
            return None

        experiment_id = active_experiment.experiment_id
        # TODO: There may be race conditions in the below code for parallel
        # steps. For example for HP tuning if two train steps are running
        # and they both create a run then we send it onwards to a testing step
        # which will now not know which run to use. We don't want a new run
        # But rather to search for the most recent run. But even that might
        # Have conflicts when there are multiple running steps that do this...
        # How can we handle this?
        # Naive Idea: How about each step has an identifier with it?
        runs = mlflow.search_runs(
            experiment_ids=[experiment_id],
            filter_string=f'tags.mlflow.runName = "{step_env.pipeline_run_id}"',
            output_format="list",
        )

        run_id = runs[0].info.run_id if runs else None

        current_active_run = mlflow.active_run()
        if current_active_run and current_active_run.info.run_id == run_id:
            # Is not None AND run_id matches
            return current_active_run

        return mlflow.start_run(
            run_id = run_id,
            run_name=step_env.pipeline_run_id,
            experiment_id=experiment_id,
        )

    def log_params(self, params: Dict[str, Any]) -> None:
        raise NotImplementedError()

    def log_metrics(self, metrics: Dict[str, Any]) -> None:
        raise NotImplementedError()

    def log_artifacts(self, artifacts: Dict[str, Any]) -> None:
        raise NotImplementedError()



import logging
from logging.handlers import TimedRotatingFileHandler
from typing import Any, Dict, List

from rich.logging import RichHandler
from rich.traceback import install as rich_tb_install

from coalescenceml.constants import APP_NAME, COML_LOGGING_VERBOSITY
from coalescenceml.enums import LoggingLevels


LOG_FILE = f"{APP_NAME}_logs.log"


class CustomFormatter(logging.Formatter):
    """Formats logs according to our specification."""

    grey: str = "\x1b[38;21m"
    pink: str = "\x1b[35m"
    green: str = "\x1b[32m"
    blue: str = "\x1b[34m"
    yellow: str = "\x1b[33m"
    red: str = "\x1b[31m"
    bold_red: str = "\x1b[31;1m"
    purple: str = "\x1b[1;35m"
    reset: str = "\x1b[0m"

    format_template: str = (
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%("
        "filename)s:%(lineno)d)"
        if LoggingLevels[COML_LOGGING_VERBOSITY] == LoggingLevels.DEBUG
        else "%(message)s"
    )

    COLORS: Dict[LoggingLevels, str] = {
        LoggingLevels.DEBUG: purple,
        LoggingLevels.INFO: blue,
        LoggingLevels.WARN: yellow,
        LoggingLevels.ERROR: red,
        LoggingLevels.CRITICAL: bold_red,
    }

    def format(self, log_record: logging.LogRecord) -> str:
        """Convert log record into a formatted string.

        Args:
            log_record: LogRecord generated by logging module

        Returns:
            A string formatted according to specifications.
        """
        log_fmt = (
            self.COLORS[LoggingLevels(log_record.levelno)]
            + self.format_template
            + self.reset
        )

        formatter = logging.Formatter(log_fmt)
        formatted_record = formatter.format(log_record)
        return formatted_record


def get_logging_level() -> LoggingLevels:
    """Get logging level from ENV.

    Returns:
        current verbosity of logger

    Raises:
        KeyError: If the verbosity stored in env / constants is not a member
            of the LoggingLevels enum
    """
    verbosity = COML_LOGGING_VERBOSITY.upper()
    if verbosity not in LoggingLevels.__members__:
        raise KeyError(
            f"Verbosity must be one of "
            f"{list(LoggingLevels.__members__.keys())} not {verbosity}"
        )
    return LoggingLevels[verbosity]


def get_console_handler() -> Any:
    """Get console handler for logging.

    Returns:
        stream handler for rich console
    """
    from coalescenceml.constants import console

    console_handler = RichHandler(
        show_path=False,
        omit_repeated_times=False,
        console=console,
        rich_tracebacks=True,
    )
    console_handler.setFormatter(CustomFormatter())
    return console_handler


def get_file_handler() -> Any:
    """Return a file handler for logging.

    Returns:
        file handler for logs
    """
    file_handler = TimedRotatingFileHandler(LOG_FILE, when="midnight")
    file_handler.setFormatter(CustomFormatter())
    return file_handler


def get_logger(logger_name: str) -> logging.Logger:
    """Fetch logger with name.

    Args:
        logger_name: Name of logger to initialize.

    Returns:
        A logger object.
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(get_logging_level().value)
    logger.addHandler(get_console_handler())

    # TODO: We hand persistent file handing and we need to double
    # logger.addHandler(get_file_handler())
    # check on the propagate issue whether this is OK
    logger.propagate = False
    return logger


def set_root_verbosity() -> None:
    """Set main verbosity of module."""
    level = get_logging_level()
    if level == LoggingLevels.NOTSET:
        logging.getLogger().disabled = True
        get_logger(__name__).debug("Logging NOTSET")
    else:
        rich_tb_install(show_locals=(level == LoggingLevels.DEBUG))

        logging.basicConfig(level=level.value)
        get_logger(__name__).debug(
            f"Logging set to level: {logging.getLevelName(level.value)}"
        )


def init_logging() -> None:
    """Initialize logging with default configuration."""
    set_root_verbosity()

    muted_loggers: List[str] = []
    for logger_name in muted_loggers:
        logging.getLogger(logger_name).setLevel(logging.WARNING)
        logging.getLogger(logger_name).disabled = True

    # ML-Metadata and ml-pipelines-sdk apparently use absl (which makes sense)
    # BUT we want to hide their logging so we will try this.
    from absl import logging as absl_logging

    absl_logging.set_verbosity(-10)



from typing import Callable, List, Optional, Sequence, Type, TypeVar, Union, overload

from coalescenceml.pipeline.base_pipeline import (
    INSTANCE_CONFIGURATION,
    PARAM_DOCKERIGNORE_FILE,
    PARAM_ENABLE_CACHE,
    PARAM_REQUIRED_INTEGRATIONS,
    PARAM_REQUIREMENTS_FILE,
    PARAM_SECRETS,
    PIPELINE_INNER_FUNC_NAME,
    BasePipeline,
)


F = TypeVar("F", bound=Callable[..., None])


@overload
def pipeline(func: F) -> Type[BasePipeline]:
    ...


@overload
def pipeline(
    *,
    name: Optional[str] = None,
    enable_cache: bool = True,
    required_integrations: Sequence[str] = (),
    requirements_file: Optional[str] = None,
    dockerignore_file: Optional[str] = None,
    secrets: Optional[List[str]] = [],
) -> Type[BasePipeline]:
    ...


def pipeline(
    func: Optional[F] = None,
    *,
    name: Optional[str] = None,
    enable_cache: bool = True,
    required_integrations: Sequence[str] = (),
    requirements_file: Optional[str] = None,
    dockerignore_file: Optional[str] = None,
    secrets: Optional[List[str]] = [],
) -> Union[Callable[[F], Type[BasePipeline]], Type[BasePipeline]]:
    """ """
    def pipe_decorator(func_: F) -> Type[BasePipeline]:
        pipeline_name = name or func_.__name__
        return type(
            pipeline_name,
            (BasePipeline,),
            {
                PIPELINE_INNER_FUNC_NAME: staticmethod(func_),
                INSTANCE_CONFIGURATION: {
                    PARAM_ENABLE_CACHE: enable_cache,
                    PARAM_REQUIRED_INTEGRATIONS: required_integrations,
                    PARAM_REQUIREMENTS_FILE: requirements_file,
                    PARAM_DOCKERIGNORE_FILE: dockerignore_file,
                    PARAM_SECRETS: secrets,
                },
            },
        )

    if func is None:
        return pipe_decorator
    else:
        return pipe_decorator(func)



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, ClassVar

from tfx.dsl.compiler.compiler import Compiler
from tfx.dsl.compiler.constants import PIPELINE_RUN_ID_PARAMETER_NAME
from tfx.dsl.components.base import base_component
from tfx.orchestration import metadata
from tfx.orchestration.local import runner_utils
from tfx.orchestration.pipeline import Pipeline as TfxPipeline
from tfx.orchestration.portable import launcher, runtime_parameter_utils
from tfx.proto.orchestration import executable_spec_pb2
from tfx.proto.orchestration.pipeline_pb2 import (
    Pipeline as Pb2Pipeline,
    PipelineNode,
)

from coalescenceml.directory import Directory
from coalescenceml.enums import MetadataContextFlavor
from coalescenceml.logger import get_logger
from coalescenceml.orchestrator import BaseOrchestrator, utils


if TYPE_CHECKING:
    from coalescenceml.pipeline.base_pipeline import BasePipeline
    from coalescenceml.pipeline.runtime_configuration import (
        RuntimeConfiguration,
    )
    from coalescenceml.stack import Stack


logger = get_logger(__name__)


class LocalOrchestrator(BaseOrchestrator):
    """Orchestrator responsible for running pipelines locally."""

    # Class Configuration
    FLAVOR: ClassVar[str] = "local"

    def run_pipeline(
        self,
        pipeline: BasePipeline,
        stack: Stack,
        runtime_configuration: RuntimeConfiguration,
    ) -> Any:
        """Runs a pipeline locally"""

        tfx_pipeline: TfxPipeline = utils.create_tfx_pipeline(
            pipeline, stack=stack
        )

        if runtime_configuration is None:
            runtime_configuration = RuntimeConfiguration()

        if runtime_configuration.schedule:
            logger.warning(
                "Local Orchestrator currently does not support the"
                "use of schedules. The `schedule` will be ignored "
                "and the pipeline will be run directly"
            )

        pipeline_root = tfx_pipeline.pipeline_info.pipeline_root
        if not isinstance(pipeline_root, str):
            raise TypeError(
                "TFX Pipeline root may not be a Placeholder, "
                "but must be a specific string."
            )

        for component in tfx_pipeline.components:
            if isinstance(component, base_component.BaseComponent):
                component._resolve_pip_dependencies(pipeline_root)

        pb2_pipeline: Pb2Pipeline = Compiler().compile(tfx_pipeline)

        # Substitute the runtime parameter to be a concrete run_id
        runtime_parameter_utils.substitute_runtime_parameter(
            pb2_pipeline,
            {
                PIPELINE_RUN_ID_PARAMETER_NAME: runtime_configuration.run_name,
            },
        )

        deployment_config = runner_utils.extract_local_deployment_config(
            pb2_pipeline
        )
        connection_config = (
            Directory().active_stack.metadata_store.get_tfx_metadata_config()
        )

        logger.debug(f"Using deployment config:\n {deployment_config}")
        logger.debug(f"Using connection config:\n {connection_config}")

        # Run each component. Note that the pipeline.components list is in
        # topological order.
        for node in pb2_pipeline.nodes:
            pipeline_node: PipelineNode = node.pipeline_node

            # fill out that context
            utils.add_context_to_node(
                pipeline_node,
                type_=MetadataContextFlavor.STACK.value,
                name=str(hash(json.dumps(stack.dict(), sort_keys=True))),
                properties=stack.dict(),
            )

            # Add all pydantic objects from runtime_configuration to the context
            utils.add_runtime_configuration_to_node(
                pipeline_node, runtime_configuration
            )

            # Add pipeline requirements as a context
            requirements = " ".join(sorted(pipeline.requirements))
            utils.add_context_to_node(
                pipeline_node,
                type_=MetadataContextFlavor.PIPELINE_REQUIREMENTS.value,
                name=str(hash(requirements)),
                properties={"pipeline_requirements": requirements},
            )

            node_id = pipeline_node.node_info.id
            executor_spec = runner_utils.extract_executor_spec(
                deployment_config, node_id
            )
            custom_driver_spec = runner_utils.extract_custom_driver_spec(
                deployment_config, node_id
            )

            p_info = pb2_pipeline.pipeline_info
            r_spec = pb2_pipeline.runtime_spec

            # set custom executor operator to allow custom execution logic for
            # each step
            step = utils.get_step_for_node(
                pipeline_node, steps=list(pipeline.steps.values())
            )
            custom_executor_operators = {
                executable_spec_pb2.PythonClassExecutableSpec: step.executor_operator
            }

            component_launcher = launcher.Launcher(
                pipeline_node=pipeline_node,
                mlmd_connection=metadata.Metadata(connection_config),
                pipeline_info=p_info,
                pipeline_runtime_spec=r_spec,
                executor_spec=executor_spec,
                custom_driver_spec=custom_driver_spec,
                custom_executor_operators=custom_executor_operators,
            )
            stack.prepare_step_run()
            utils.execute_step(component_launcher)
            stack.cleanup_step_run()



import os
import tempfile
from typing import Any, Type

import xgboost as xgb

from coalescenceml.artifacts import ModelArtifact
from coalescenceml.io import fileio
from coalescenceml.producers import BaseProducer
from coalescenceml.producers.producer_registry import register_producer_class


DEFAULT_FILENAME = "model.xgb.json"


@register_producer_class
class XgboostBoosterProducer(BaseProducer):
    """Producer to read and write from xgboost.Booster."""

    TYPES = (xgb.Booster,)
    ARTIFACT_TYPES = (ModelArtifact,)

    def handle_input(self, data_type: Type[Any]) -> xgb.Booster:
        """Read XGB Booster model from JSON file.

        Args:
            data_type: Data type to be processed.

        Returns:
            XGB Booster model
        """
        super().handle_input(data_type)
        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)

        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = os.path.join(temp_dir, DEFAULT_FILENAME)
            # Create temp file
            # Copy from artifact store to temp file
            fileio.copy(filepath, temp_file)
            booster = xgb.Booster()
            booster.load_model(temp_file)

        return booster

    def handle_return(self, booster: xgb.Booster) -> None:
        """Create JSON file for Booster.

        Args:
            booster: An xgboost Booster model
        """
        super().handle_return(booster)

        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)

        # Make temp artifact
        with tempfile.NamedTemporaryFile(suffix=".json", mode="w", delete=True) as temp_file:
            booster.save_model(temp_file.name)

            # copy file
            fileio.copy(temp_file.name, filepath)



import os
from contextlib import contextmanager
from typing import ClassVar, Iterator, Optional, Tuple

import wandb

from coalescenceml.integrations.contants import WANDB
from coalescenceml.logger import get_logger
from coalescenceml.stack.stack_component_registry import (
    register_stack_component_class,
)

logger = get_logger(__name__)

WANDB_API_KEY = "WANDB_API_KEY"


@register_stack_component_class
class WandbExperimentTracker(BaseExperimentTracker):
    """Store wandb configuration and execute logging.

    ..note: One method of directly accessing the experiment
        tracker is as follows with the StepContext....

    ```python
    from coalescenceml.step import StepContext


    @enable_wandb
    @step
    def my_train_step(context: StepContext, ...):
        context.stack.experiment_tracker # Get experiment tracker directly
    ```

    ..note: Otherwise directly using wand.log({"acc": 0.5}) will work
        perfectly!

    # TODO: Determine an easy way to access the experiment tracker and
    # the corresponding helper functions without accessing the context.
    # This might be an OK pattern but might be tough to remember
    """

    api_key: str
    entity: Optional[str] = None
    project_name: Optional[str] = None

    FLAVOR: ClassVar[str] = WANDB

    def prepare_step_run(self) -> None:
        """Set the wandb API key."""
        os.environ[WANDB_API_KEY] = self.api_key

    def cleanup_step_run(self) -> None:
        """Clean up wandb API key."""
        os.environ.pop(WANDB_API_KEY, None)

    @contextmanager
    def with_wandb_run(
        self,
        run_name: str,
        tags: Tuple[str, ...] = (),
        settings: Optional[wandb.Settings] = None,
    ) -> Iterator[None]:
        """Activate wandb run for duration of context manager.

        Anything logged to wandb that is run while this is active
        will automatically log to the same wandb run configured
        by the run name passed here.
  
        Args:
            run_name: Name of the wandb run to create
            tags: tags to attach to wandb run
            settings: Additional settings for the wandb run
        """
        try:
            logger.info(
                f"Initializing wandb with project name: {self.project_name} "
                f"run name: {run_name} entity: {self.entity}"
            )
            wandb.init(
                project=self.project_name,
                name=run_name,
                entity=self.entity,
                settings=settings,
                tags=tags,
            )
            yield
        finally:
            wandb.finish()

    def log_params(self, params: Dict[str, Any]) -> None:
        wandb.config.update(params)

    def log_metrics(self, metrics: Dict[str, Any]) -> None:
        wandb.log(metrics)

    def log_artifacts(self, artifacts: Dict[str, Any]) -> None:
        raise NotImplementedError()



from typing import Any, Type
import os
import pickle

import statsmodels as sm
from statsmodels.base.wrapper import ResultsWrapper

from coalescenceml.artifacts import ModelArtifact
from coalescenceml.io import fileio
from coalescenceml.logger import get_logger
from coalescenceml.producers import BaseProducer
from coalescenceml.producers.producer_registry import register_producer_class

logger = get_logger(__name__)
DEFAULT_FILENAME = "model.pkl" # TODO: Can the users specify a file name?


@register_producer_class
class StatsmodelsProducer(BaseProducer):
    """Producer to handle statsmodels model classes"""
    
    ARTIFACT_TYPES = (ModelArtifact,)
    TYPES = (ResultsWrapper,)

    def handle_input(self, data_type: Type[Any]) -> ResultsWrapper:
        """Process artifact of data_type into a Statsmodel object.

        Args:
            data_type: data type to check and process

        Returns:
            ResultsWrapper from statsmodels
        """
        super().handle_input(data_type)

        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)
        with fileio.open(filepath, "rb") as fp:
            contents = pickle.load(fp) # TODO: Can we make this more secure somehow?

        return contents

    def handle_return(self, model: ResultsWrapper) -> None:
        """Write statsmodels model to artifact store as pickle.

        Args:
            model: Statsmodels model to save.
        """
        super().handle_return(model)

        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)
        with fileio.open(filepath, "wb") as fp:
            pickle.dump(model, fp)



import json
import os
from typing import AbstractSet, Any, Dict, Iterable, List, Optional, cast

import pkg_resources
from docker.client import DockerClient
from docker.utils import build as docker_build

import coalescenceml
from coalescenceml import __version__
from coalescenceml.config.global_config import GlobalConfiguration
from coalescenceml.constants import ENV_COML_CONFIG_PATH
from coalescenceml.io import fileio
from coalescenceml.io.utils import read_file_contents_as_string
from coalescenceml.logger import get_logger
from coalescenceml.utils import readability_utils

DEFAULT_BASE_IMAGE = "python:3.8.13-slim-buster"
CUDA_BASE_IMAGE = "tensorflow/tensorflow:2.6.1-gpu"
CONTAINER_COML_CONFIG_DIR = ".coalescenceconfig"

logger = get_logger(__name__)


def parse_dockerignore(dockerignore_path: str) -> List[str]:
    """Parse a dockerignore and return list of patterns.

    Args:
        dockerignore_path: location of dockerignore

    Returns:
        list of patterns to match
    """
    try:
        file_content = read_file_contents_as_string(dockerignore_path)
    except FileNotFoundError:
        logger.warning(
            f"Unable to find dockerignore file at path '{dockerignore_path}'"
        )
        return []

    exclude_patterns = []
    for line in file_content.split("\n"):
        line = line.strip()
        if line and not line.startswith("#"):
            exclude_patterns.append(line)

    return exclude_patterns


def generate_dockerfile_contents(
    base_image: str,
    entrypoint: Optional[str] = None,
    requirements: Optional[AbstractSet[str]]= None,
    environment_vars: Optional[Dict[str, str]] = None,
) -> str:
    """Generate a Dockerfile.

    Args:
        base_image: The image to use as a base for the dockerfile.
        entrypoint: The default entrypoint command that is to be executed
            when running the container.
        requirements: List of pip requirements to install.
        environment_vars: Dict of environment variables to set.

    Returns:
        contents of Dockerfile
    """
    lines = [f"from {base_image}", "WORKDIR /app"]

    if environment_vars:
        for key, value in environment_vars.items():
            lines.append(f"ENV {key}={value}")

    lines.append(f"RUN pip install --no-cache coalescenceml=={__version__}")
    if requirements:
        lines.append(
            f"RUN pip install --no-cache {' '.join(sorted(requirements))}"
        )

    lines.append("COPY . .")
    lines.append("RUN chmod -R a+rw .")
    lines.append(f"ENV {ENV_COML_CONFIG_PATH}=/app/{CONTAINER_COML_CONFIG_DIR}")

    if entrypoint:
        lines.append(f"ENTRYPOINT {entrypoint}")

    return "\n".join(lines)


def create_build_context(
    build_context_path: str,
    dockerfile_contents: str,
    dockerignore_path: Optional[str] = None,
) -> Any:
    default_dockerignore_path = os.path.join(
        build_context_path, ".dockerignore"
    )

    if dockerignore_path:
        exclude_patterns = parse_dockerignore(dockerignore_path)
    elif fileio.exists(default_dockerignore_path):
        exclude_patterns = parse_dockerignore(default_dockerignore_path)
    else:
        exclude_patterns = []


    logger.debug(
        f"Exclude patterns for docker build context: {exclude_patterns}"
    )

    files = docker_build.exclude_paths(
        build_context_path, patterns=exclude_patterns
    )

    context = docker_build.create_archive(
        root=build_context_path,
        files=sorted(files),
        gzip=False,
        extra_files=[
            ("Dockerfile", dockerfile_contents),
        ]
    )

    build_context_size = os.path.getsize(context.name)
    filesize_50mb = (1024 * 1024) * 50
    if build_context_size > filesize_50mb:
        # Build context exceeds 50mb; Make sure to ignore things!
        logger.warning(
            f"Build context size for docker image: "
            f"{readability_utils.get_human_readable_filesize}. If you believe "
            f"this is large, make sure to include a `.dockerignore` file at "
            f"the root of your build context '{build_context_path}' or specify "
            f"a custom file when defining your pipeline."
        )

    return context


def get_current_environment_requirements() -> Dict[str, str]:
    """Return package requirements for current python process.

    Returns:
        A dict of package requirements for current environment
    """
    return {
        distribution.key: distribution.version
        for distribution in pkg_resources.working_set
    }

def process_docker_stream(stream: Iterable[bytes]) -> None:
    """Process output stream of docker command.

    Raises:
        JSONDecodeError: If an element of the stream is not json decodable.
        RuntimeError: If there was an error while running the docker command.
    """
    for element in stream:
        lines = element.decode("utf-8").strip().split("\n")

        for line in lines:
            try:
                line_json = json.loads(line)
                if "error" in line_json:
                    raise RuntimeError(f"Docker Error: {line_json['error']}")
                elif "stream" in line_json:
                    logger.info(line_json['stream'].strip())
                else:
                    pass
            except json.JSONDecodeError as error:
                logger.warning(
                    f"Failed to decode json for line '{line}' with "
                    f"error: {error}"
                )


def push_docker_image(image_name: str) -> None:
    """Push a docker image to registry.

    Args:
        image_name: The name and tag of the image to push.
    """
    logger.info(f"Pushing docker image '{image_name}'")
    docker_client = DockerClient.from_env()
    output_stream = docker_client.images.push(image_name, stream=True)
    process_docker_stream(output_stream)


def build_docker_image(
    build_context_path: str,
    image_name: str,
    entrypoint: Optional[str] =None,
    dockerfile_path: Optional[str] = None,
    dockerignore_path: Optional[str] = None,
    requirements: Optional[AbstractSet[str]] = None,
    environment_vars: Optional[Dict[str, str]] = None,
    use_local_requirements: bool =False,
    base_image: Optional[str] = None,
) -> None:
    """Build a docker image.

    Args:
    """
    config_path = os.path.join(build_context_path, CONTAINER_COML_CONFIG_DIR)
    try:
        # TODO: Find a cleaner way than copying the literal profiles and
        # stack into the docker container. If we happen to store secrets
        # then those will be stored in our container....
        # Save global config with active profile and stack in build context
        # so that they are accessible within the container
        GlobalConfiguration().copy_active_configuration(
            config_path,
            load_config_path=f"/app/{CONTAINER_COML_CONFIG_DIR}",
        )

        if not requirements and use_local_requirements:
            local_requirements = get_current_environment_requirements()
            requirements = {
                f"{package}=={version}"
                for package, version in local_requirements.items()
                if package != "coalescenceml"
            }
            logger.info(
                f"Using requirements from local environment for docker "
                f"image: {requirements}"
            )
        elif requirements and use_local_requirements:
            logger.warning(
                "Using requirements passed in and local requirements. "
                "This has less defined behavior and may be changed in the "
                "future. There are no guarantees your dependencies resolve."
            )
            local_requirements = get_current_environment_requirements()
            requirements = requirements.union(
                {
                    f"{package}=={version}"
                    for package, version in local_requirements.items()
                    if package != "coalescenceml"
                }
            )

        if dockerfile_path:
            dockerfile_contents = read_file_contents_as_string(dockerfile_path)
        else:
            dockerfile_contents = generate_dockerfile_contents(
                base_image=base_image or DEFAULT_BASE_IMAGE,
                entrypoint=entrypoint,
                requirements=requirements,
                environment_vars=environment_vars,
            )

        build_context = create_build_context(
            build_context_path=build_context_path,
            dockerfile_contents=dockerfile_contents,
            dockerignore_path=dockerignore_path,
        )

        logger.info(
            f"Building docker image '{image_name}'"
        )

        docker_client = DockerClient.from_env()
        output_stream = docker_client.images.client.api.build(
            fileobj=build_context,
            custom_context=True,
            tag=image_name,
        )

        process_docker_stream(output_stream)
    finally:
        # Clean temporary build files
        fileio.rmtree(config_path)

    logger.info("Finished building docker image.")

def get_image_digest(image_name: str) -> Optional[str]:
    docker_client = DockerClient.from_env()
    image = docker_client.images.get(image_name)
    repo_digests = image.attrs["RepoDigests"]
    if len(repo_digests) == 1:
        return cast(str, repo_digests[0])
    else:
        logger.debug(
            "Found 0 or 2+ repo digests for the image "
            f"'{image_name}': {repo_digests}"
        )



from coalescenceml.integrations.constants import AZURE
from coalescenceml.integrations.integration import Integration


class AzureIntegration(Integration):
    """Integration for AZURE."""

    NAME = AZURE
    REQUIREMENTS = ["adlfs==2022.4.0"]

    @classmethod
    def activate(cls) -> None:
        from coalescenceml.integrations.azure_datalake import artifact_store


AzureIntegration.check_installation()


"""
placeholder init message
"""
from coalescenceml.integrations.constants import SKLEARN
from coalescenceml.integrations.integration import Integration

class SKLearnIntegration(Integration):
    """Integration class for SKLearn"""

    NAME = SKLEARN

    REQUIREMENTS = ["scikit-learn==1.0.2"]

    @classmethod
    def activate(cls) -> None:
        from coalescenceml.integrations.sklearn import producers

SKLearnIntegration.check_installation()



"""
## Stack
The stack is essentially all the configuration for the infrastructure of your 
MLOps platform.
A stack is made up of multiple components. Some examples are:
- An Artifact Store
- A Metadata Store
- An Orchestrator
- A Container Registry (Optional)
"""

from coalescenceml.stack.stack import Stack
from coalescenceml.stack.stack_component import StackComponent
from coalescenceml.stack.stack_validator import StackValidator


__all__ = [
    "Stack",
    "StackComponent",
    "StackValidator",
]



from pathlib import Path
from shutil import rmtree
from typing import Optional

import click

from coalescenceml.cli.cli import cli
from coalescenceml.cli.utils import confirmation, error, info
from coalescenceml.config.global_config import GlobalConfiguration
from coalescenceml.constants import DIRECTORY_DIRECTORY_NAME, console
from coalescenceml.directory import Directory
from coalescenceml.exceptions import InitializationException


@cli.command("init", help="Initialize a CoalescenceML directory.")
@click.option(
    "--path",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=Path
    ),
)
def init(path: Optional[Path]) -> None:
    """Initialize CoalescenceML on given path.
    Args:
      path: Path to the directory.
    Raises:
        InitializationException: If the repo is already initialized.
    """
    if path is None:
        path = Path.cwd()

    with console.status(f"Initializing CoalescenceML directory at {path}.\n"):
        try:
            Directory.initialize(root=path)
            info(f"CoalescenceML directory initialized at {path}.")
        except InitializationException as e:
            error(f"{e}")

    cfg = GlobalConfiguration()
    info(
        f"The local active profile was initialized to "
        f"'{cfg.active_profile_name}' and the local active stack to "
        f"'{cfg.active_stack_name}'. This local configuration will only take "
        f"effect when you're running CoalescenceML from the initialized directory "
        f"root, or from a subdirectory."
    )


# TODO: Make this actually clean the stuff LMAO.
@cli.command("clean")
@click.option("--yes", "-y", type=click.BOOL, default=False)
def clean(yes: bool = False) -> None:
    """Clean everything in directory.

    Args:
      yes: bool:  (Default value = False)
    """
    if not yes:
        yes = confirmation(
            "This will completely delete all pipelines, their associated "
            "artifacts and metadata ever created in this CoalescenceML directory."
            "Are you sure you want to proceed?"
        )

    if yes:
        dir_path = (
            Path(Directory().get_active_root()) / DIRECTORY_DIRECTORY_NAME
        )
        rmtree(dir_path)
    else:
        info("Skipping deletion of CoalescenceML directory...")



from __future__ import annotations

import os
from typing import Any, Type

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
from numpy.typing import NDArray

from coalescenceml.artifacts import DataArtifact
from coalescenceml.io import fileio
from coalescenceml.producers.base_producer import BaseProducer
from coalescenceml.producers.producer_registry import register_producer_class
from coalescenceml.utils.json_utils import read_json, write_json


DATA_FILENAME = "data.parquet"
SHAPE_FILENAME = "shape.json"
DATA_VAR = "data_var"


@register_producer_class
class NumpyProducer(BaseProducer):
    """Producer to read data to and from numpy arrays."""

    ARTIFACT_TYPES = (DataArtifact,)
    TYPES = (np.ndarray,)

    def handle_input(self, data_type: Type[Any]) -> NDArray[Any]:
        """Reads numpy array from parquet file.

        Args:
            data_type: type of input to be processed

        Returns:
            Numpy array with data
        """
        super().handle_input(data_type)

        shape_dict = read_json(os.path.join(self.artifact.uri, SHAPE_FILENAME))
        shape_tuple = tuple(shape_dict.values())

        with fileio.open(
            os.path.join(self.artifact.uri, DATA_FILENAME), "rb"
        ) as fp:
            input_stream = pa.input_stream(fp)
            data = pq.read_table(input_stream)

        vals = getattr(data.to_pandas(), DATA_VAR).values
        return np.reshape(vals, shape_tuple)

    def handle_return(self, arr: NDarray[Any]) -> None:
        """Writes a np.ndarray to artifact store as parquet.

        Args:
            arr: Numpy arry to write.
        """
        super().handle_return(arr)

        write_json(
            os.path.join(self.artifact.uri, SHAPE_FILENAME),
            {str(i): d for i, d in enumerate(arr.shape)},
        )

        pa_table = pa.table({DATA_VAR: arr.flatten()})
        with fileio.open(
            os.path.join(self.artifact.uri, DATA_FILENAME), "wb"
        ) as fp:
            stream = pa.output_stream(fp)
            pq.write_table(pa_table, stream)


# producer_registry.register_producer(np.ndarray, NumpyProducer)



"""
placeholder init message
"""
from coalescenceml.integrations.constants import TENSORFLOW
from coalescenceml.integrations.integration import Integration

class TFIntegration(Integration):
    """Integration class for TensorFlow"""

    NAME = TENSORFLOW

    REQUIREMENTS = ["tensorflow==2.8.0", "tensorflow_io==0.24.0"]

    @classmethod
    def activate(cls) -> None:
        from coalescenceml.integrations.tensorflow import producers

TFIntegration.check_installation()



from typing import Optional, Tuple

import click
from rich.markdown import Markdown
from rich.progress import track

from coalescenceml.cli.cli import cli
from coalescenceml.cli.utils import (
    confirmation,
    error,
    format_integration_list,
    info,
    install_packages,
    print_table,
    title,
    uninstall_package,
    warning,
)
from coalescenceml.constants import console
from coalescenceml.integrations.registry import integration_registry
from coalescenceml.logger import get_logger


logger = get_logger(__name__)


@cli.group(help="Interact with the requirements of external integrations.")
def integration() -> None:
    """Integrations group"""


@integration.command(name="list", help="List the available integrations.")
def list_integrations() -> None:
    """List all available integrations with their installation status."""
    formatted_table = format_integration_list(
        [x for x in integration_registry.integrations.items()]
    )
    print_table(formatted_table)
    warning(
        "\n" + "To install the dependencies of a specific integration, type: "
    )
    warning("coml integration install EXAMPLE_NAME")


@integration.command(
    name="requirements", help="List all requirements for an integration."
)
@click.argument("integration_name", required=False, default=None)
def get_requirements(integration_name: Optional[str] = None) -> None:
    """List all requirements for the chosen integration."""
    try:
        requirements = integration_registry.select_integration_requirements(
            integration_name
        )
    except KeyError as e:
        error(str(e))
    else:
        if requirements:
            title(
                f'Requirements for {integration_name or "all integrations"}:\n'
            )
            info(f"{requirements}")
            warning(
                "\n" + "To install the dependencies of a "
                "specific integration, type: "
            )
            warning("coml integration install EXAMPLE_NAME")


# TODO: How do we handle system requirements?
@integration.command(
    help="Install the required packages for the integration of choice."
)
@click.argument("integrations", nargs=-1, required=False)
@click.option(
    "--ignore-integration",
    "-i",
    multiple=True,
    help="List of integrations to ignore explicitly.",
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force the installation of the required packages. This will skip the "
    "confirmation step and reinstall existing packages as well",
)
def install(
    integrations: Tuple[str],
    ignore_integration: Tuple[str],
    force: bool = False,
) -> None:
    """Installs the required packages for a given integration. If no integration
    is specified all required packages for all integrations are installed
    using pip"""
    if not integrations:
        # no integrations specified, use all registered integrations
        integrations = set(integration_registry.integrations.keys())
        for i in ignore_integration:
            try:
                integrations.remove(i)
            except KeyError:
                error(
                    f"Integration {i} does not exist. Available integrations: "
                    f"{list(integration_registry.integrations.keys())}"
                )

    requirements = []
    integrations_to_install = []
    for integration_name in integrations:
        try:
            if force or not integration_registry.is_installed(integration_name):
                requirements += (
                    integration_registry.select_integration_requirements(
                        integration_name
                    )
                )
                integrations_to_install.append(integration_name)
            else:
                info(
                    f"All required packages for integration "
                    f"'{integration_name}' are already installed."
                )
        except KeyError:
            warning(f"Unable to find integration '{integration_name}'.")

    if requirements and (
        force
        or confirmation(
            "Are you sure you want to install the following "
            "packages to the current environment?\n"
            f"{requirements}"
        )
    ):
        with console.status("Installing integrations..."):
            install_packages(requirements)


@integration.command(
    help="Uninstall the required packages for the integration of choice."
)
@click.argument("integrations", nargs=-1, required=False)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force the uninstallation of the required packages. This will skip "
    "the confirmation step",
)
def uninstall(integrations: Tuple[str], force: bool = False) -> None:
    """Installs the required packages for a given integration. If no integration
    is specified all required packages for all integrations are installed
    using pip"""
    if not integrations:
        # no integrations specified, use all registered integrations
        integrations = tuple(integration_registry.integrations.keys())

    requirements = []
    for integration_name in integrations:
        try:
            if integration_registry.is_installed(integration_name):
                requirements += (
                    integration_registry.select_integration_requirements(
                        integration_name
                    )
                )
            else:
                warning(
                    f"Requirements for integration '{integration_name}' "
                    f"already not installed."
                )
        except KeyError:
            warning(f"Unable to find integration '{integration_name}'.")

    if requirements and (
        force
        or confirmation(
            "Are you sure you want to uninstall the following "
            "packages from the current environment?\n"
            f"{requirements}"
        )
    ):
        for n in track(
            range(len(requirements)),
            description="Uninstalling integrations...",
        ):
            uninstall_package(requirements[n])


@integration.command("explain")
def explain_integrations() -> None:
    """Explains the concept of CoalescenceML integrations."""

    with console.pager():
        console.print(
            Markdown(
                """
TODO: Explain in words and with examples about what an integration is and the CLI
command. This should showcase one of every command probably and the most common
methods that they will want to use.
"""
            )
        )



import json
from typing import Dict, Any, Optional

import click

import coalescenceml
from coalescenceml.cli import utils as cli_utils
from coalescenceml.cli.cli import cli
from coalescenceml.cli.stack_components import (
    component_display_name_helper,
    register_stack_component_helper,
)
from coalescenceml.config.global_config import GlobalConfiguration
from coalescenceml.constants import console
from coalescenceml.directory import Directory
from coalescenceml.enums import StackComponentFlavor
from coalescenceml.stack import Stack
from coalescenceml.stack.exceptions import ProvisioningError
from coalescenceml.stack.stack_component_class_registry import (
    StackComponentClassRegistry,
)
from coalescenceml.utils import yaml_utils


# Stacks
@cli.group()
def stack() -> None:
    """Stacks to define various environments."""


@stack.command("register", context_settings=dict(ignore_unknown_options=True))
@click.argument("stack_name", type=str, required=True)
@click.option(
    "-m",
    "--metadata-store",
    "metadata_store_name",
    help="Name of the metadata store for this stack.",
    type=str,
    required=True,
)
@click.option(
    "-a",
    "--artifact-store",
    "artifact_store_name",
    help="Name of the artifact store for this stack.",
    type=str,
    required=True,
)
@click.option(
    "-o",
    "--orchestrator",
    "orchestrator_name",
    help="Name of the orchestrator for this stack.",
    type=str,
    required=True,
)
@click.option(
    "-c",
    "--container_registry",
    "container_registry_name",
    help="Name of the container registry for this stack.",
    type=str,
    required=False,
)
@click.option(
    "-x",
    "--secrets_manager",
    "secrets_manager_name",
    help="Name of the secrets manager for this stack.",
    type=str,
    required=False,
)
@click.option(
    "-s",
    "--step_operator",
    "step_operator_name",
    help="Name of the step operator for this stack.",
    type=str,
    required=False,
)
@click.option(
    "-f",
    "--feature_store",
    "feature_store_name",
    help="Name of the feature store for this stack.",
    type=str,
    required=False,
)
@click.option(
    "-d",
    "--model_deployer",
    "model_deployer_name",
    help="Name of the model deployer for this stack.",
    type=str,
    required=False,
)
@click.option(
    "-e",
    "--experiment-tracker",
    "experiment_tracker_name",
    help="Name of the experiment tracker for this stack.",
    type=str,
    required=False,
)
def register_stack(
    stack_name: str,
    metadata_store_name: str,
    artifact_store_name: str,
    orchestrator_name: str,
    container_registry_name: Optional[str] = None,
    secrets_manager_name: Optional[str] = None,
    step_operator_name: Optional[str] = None,
    feature_store_name: Optional[str] = None,
    model_deployer_name: Optional[str] = None,
    experiment_tracker_name: Optional[str] = None,
) -> None:
    """Register a stack."""
    register_stack_helper(
        stack_name=stack_name,
        metadata_store_name=metadata_store_name,
        artifact_store_name=artifact_store_name,
        orchestrator_name=orchestrator_name,
        container_registry_name=container_registry_name,
        secrets_manager_name=secrets_manager_name,
        step_operator_name=step_operator_name,
        feature_store_name=feature_store_name,
        model_deployer_name=model_deployer_name,
        experiment_tracker_name=experiment_tracker_name,
    )


def register_stack_helper(
    stack_name: str,
    metadata_store_name: str,
    artifact_store_name: str,
    orchestrator_name: str,
    container_registry_name: Optional[str] = None,
    secrets_manager_name: Optional[str] = None,
    step_operator_name: Optional[str] = None,
    feature_store_name: Optional[str] = None,
    model_deployer_name: Optional[str] = None,
    experiment_tracker_name: Optional[str] = None,
) -> None:
    cli_utils.print_active_profile()

    with console.status(f"Registering stack '{stack_name}'...\n"):
        dir_ = Directory()

        stack_components = {
            StackComponentFlavor.METADATA_STORE: dir_.get_stack_component(
                StackComponentFlavor.METADATA_STORE, name=metadata_store_name
            ),
            StackComponentFlavor.ARTIFACT_STORE: dir_.get_stack_component(
                StackComponentFlavor.ARTIFACT_STORE, name=artifact_store_name
            ),
            StackComponentFlavor.ORCHESTRATOR: dir_.get_stack_component(
                StackComponentFlavor.ORCHESTRATOR, name=orchestrator_name
            ),
        }

        if container_registry_name:
            stack_components[
                StackComponentFlavor.CONTAINER_REGISTRY
            ] = dir_.get_stack_component(
                StackComponentFlavor.CONTAINER_REGISTRY,
                name=container_registry_name,
            )

        if secrets_manager_name:
            stack_components[
                StackComponentFlavor.SECRETS_MANAGER
            ] = dir_.get_stack_component(
                StackComponentFlavor.SECRETS_MANAGER,
                name=secrets_manager_name,
            )

        if step_operator_name:
            stack_components[
                StackComponentFlavor.STEP_OPERATOR
            ] = dir_.get_stack_component(
                StackComponentFlavor.STEP_OPERATOR,
                name=step_operator_name,
            )

        if feature_store_name:
            stack_components[
                StackComponentFlavor.FEATURE_STORE
            ] = dir_.get_stack_component(
                StackComponentFlavor.FEATURE_STORE,
                name=feature_store_name,
            )
        if model_deployer_name:
            stack_components[
                StackComponentFlavor.MODEL_DEPLOYER
            ] = dir_.get_stack_component(
                StackComponentFlavor.MODEL_DEPLOYER,
                name=model_deployer_name,
            )

        if experiment_tracker_name:
            stack_components[
                StackComponentFlavor.EXPERIMENT_TRACKER
            ] = dir_.get_stack_component(
                StackComponentFlavor.EXPERIMENT_TRACKER,
                name=experiment_tracker_name,
            )

        stack_ = Stack.from_components(
            name=stack_name, components=stack_components
        )
        dir_.register_stack(stack_)
        cli_utils.info(f"Stack '{stack_name}' successfully registered!")


@stack.command("list")
def list_stacks() -> None:
    """List all available stacks in the active profile."""
    cli_utils.print_active_profile()

    dir_ = Directory()

    if len(dir_.stack_configurations) == 0:
        cli_utils.warning("No stacks registered!")
        return

    active_stack_name = dir_.active_stack_name

    stack_dicts = []
    for stack_name, stack_configuration in dir_.stack_configurations.items():
        is_active = stack_name == active_stack_name
        stack_config = {
            "ACTIVE": ":point_right:" if is_active else "",
            "STACK NAME": stack_name,
            **{
                component_type.value.upper(): value
                for component_type, value in stack_configuration.items()
            },
        }
        stack_dicts.append(stack_config)

    cli_utils.print_table(stack_dicts)


@stack.command(
    "describe",
    help="Show details about the current active stack.",
)
@click.argument(
    "stack_name",
    type=click.STRING,
    required=False,
)
def describe_stack(stack_name: Optional[str]) -> None:
    """Show details about a named stack or the active stack."""
    cli_utils.print_active_profile()

    dir_ = Directory()

    active_stack_name = dir_.active_stack_name
    stack_name = stack_name or active_stack_name

    if not stack_name:
        cli_utils.warning("No stack is set as active!")
        return

    stack_configurations = dir_.stack_configurations
    if len(stack_configurations) == 0:
        cli_utils.warning("No stacks registered!")
        return

    try:
        stack_configuration = stack_configurations[stack_name]
    except KeyError:
        cli_utils.error(f"Stack '{stack_name}' does not exist.")
        return

    cli_utils.print_stack_configuration(
        stack_configuration,
        active=stack_name == active_stack_name,
        stack_name=stack_name,
    )


@stack.command("delete")
@click.argument("stack_name", type=str)
def delete_stack(stack_name: str) -> None:
    """Delete a stack."""
    cli_utils.print_active_profile()

    with console.status(f"Deleting stack '{stack_name}'...\n"):

        cfg = GlobalConfiguration()
        dir_ = Directory()

        if cfg.active_stack_name == stack_name:
            cli_utils.error(
                f"Stack {stack_name} cannot be deleted while it is globally "
                f"active. Please choose a different active global stack first "
                f"by running 'coml stack set --global STACK'."
            )
            return

        if dir_.active_stack_name == stack_name:
            cli_utils.error(
                f"Stack {stack_name} cannot be deleted while it is "
                f"active. Please choose a different active stack first by "
                f"running 'coml stack set STACK'."
            )
            return

        Directory().deregister_stack(stack_name)
        cli_utils.info(f"Deleted stack '{stack_name}'.")


@stack.command("set")
@click.argument("stack_name", type=str)
@click.option(
    "--global",
    "-g",
    "global_profile",
    is_flag=True,
    help="Set the active stack globally.",
)
def set_active_stack(stack_name: str, global_profile: bool = False) -> None:
    """Sets a stack as active.
    If the '--global' flag is set, the global active stack will be set,
    otherwise the directory active stack takes precedence.
    """
    cli_utils.print_active_profile()

    scope = " global" if global_profile else ""

    dir_ = Directory()

    with console.status(
        f"Setting the{scope} active stack to '{stack_name}'..."
    ):

        if global_profile:
            dir_.active_profile.activate_stack(stack_name)
        else:
            dir_.activate_stack(stack_name)

        cli_utils.info(f"Active{scope} stack set to: '{stack_name}'")


@stack.command("get")
def get_active_stack() -> None:
    """Gets the active stack."""
    cli_utils.print_active_profile()

    with console.status("Getting the active stack..."):

        dir_ = Directory()
        cli_utils.info(f"The active stack is: '{dir_.active_stack_name}'")


@stack.command("up")
def up_stack() -> None:
    """Provisions resources for the active stack."""
    cli_utils.print_active_profile()

    stack_ = Directory().active_stack
    cli_utils.info(f"Provisioning resources for active stack '{stack_.name}'.")
    try:
        stack_.provision()
        stack_.resume()
    except ProvisioningError as e:
        cli_utils.error(str(e))


@stack.command("down")
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Deprovisions local resources instead of suspending them.",
)
def down_stack(force: bool = False) -> None:
    """Suspends resources of the active stack deployment."""
    cli_utils.print_active_profile()

    stack_ = Directory().active_stack

    if force:
        cli_utils.info(
            f"Deprovisioning resources for active stack '{stack_.name}'."
        )
        stack_.deprovision()
    else:
        cli_utils.info(
            f"Suspending resources for active stack '{stack_.name}'."
        )
        stack_.suspend()


def fetch_component_dict(component_flavor: StackComponentFlavor, component_name: str) -> Dict[str, str]:
    """"""
    fetched_component = Directory().get_stack_component(
        component_flavor,
        name=component_name,
    )
    component_dict = json.loads(fetched_component.json())
    tmp = {
        key: value
        for key, value in component_dict.items()
        if key != "uuid" and value is not None
    }
    tmp["flavor"] = fetched_component.FLAVOR
    return tmp


@stack.command("export")
@click.argument("stack_name", type=str, required=True)
@click.argument(
    "filepath",
    type=str,
    required=False,
)
def export_stack(stack_name: Optional[str], filepath: Optional[str]) -> None:
    """Export a stack to a YAML config."""

    # TODO: Dupliate of describe()
    dir_ = Directory()
    stack_configurations = dir_.stack_configurations
    if len(stack_configurations) == 0:
        cli_utils.error("No stacks registered at all! :((")

    try:
        stack_configuration = stack_configurations[stack_name]
    except KeyError:
        cli_utils.error(f"Stack '{stack_name}' does not exist.")

    # Create dict on stack configuration
    component_data = {}
    for component_type, component_name in stack_configuration.items():
        component_dict = fetch_component_dict(component_type, component_name)
        component_data[str(component_type)] = component_dict

    # Write out coalescence info as well to YAML
    yaml_data = {
        "coalescenceml_version": coalescenceml.__version__,
        "stack_name": stack_name,
        "components": component_data,
    }

    filepath = filepath or f"{stack_name}-config.yaml"
    yaml_utils.write_yaml(filepath, yaml_data)
    cli_utils.info(f"Exported stack '{stack_name}' to file '{filepath}'")


def import_component_helper(
    component_flavor: StackComponentFlavor, component_config: Dict[str, str],
) -> str:
    """ Return the name of a component if it exists """
    component_type = StackComponentFlavor(component_flavor)
    component_name = component_config.pop("name")
    component_flavor = component_config.pop("flavor")

    while True:
        # Check if component exists
        try:
            fetched_component_dict = fetch_component_dict(
                component_type,
                component_name,
            )

        except KeyError:
            # Component doesn't exist so it can be made
            break

        # Check if attributes are the same.
        found_component = True
        for key, value in component_config.items():
            if key not in fetched_component_dict or fetched_component_dict[key] != value:
                found_component = False
                break

        # Found existing component so return that
        if found_component:
            return component_name

        # Component exists so must be renamed
        display_name = component_display_name_helper(
            component_type
        )
        component_name = click.prompt(
            f"A '{display_name}' component with the name '{component_name}' "
            f"already exists but has a different configuration. Please "
            f"choose a different name",
            type=str,
        )

    register_stack_component_helper(
        component_type=component_type,
        component_name=component_name,
        component_flavor=component_flavor,
        **component_config
    )

    return component_name


@stack.command("import")
@click.argument("filepath", type=str, required=True)
def import_stack(filepath: str) -> None:
    """Import stack from YAML."""
    yaml_data = yaml_utils.read_yaml(filepath)
    stack_name = yaml_data["stack_name"]

    # Check COML version
    if yaml_data["coalescenceml_version"] != coalescenceml.__version__:
        cli_utils.error(
            f"Cannot import stacks from other CoalescenceML versions. "
            f"This stack was created using version {yaml_data['coalescenceml_version']} "
            f"while the current version of CoalescenceML is {coalescenceml.__version__}."
        )
        return

    dir_ = Directory()
    registered_stacks = {
        stack_.name
        for stack_ in dir_.stacks
    }
    while stack_name in registered_stacks:
        stack_name = click.prompt(
            f"Stack `{stack_name}` already exists. "
            f"Please choose a different name.",
            type=str,
        )

    # TODO: Check if any of these components are registered. If so, then
    # Skip the check. However if they have the same name but different config
    # then we should raise a helpful error message.
    # register components and stack
    component_data = {}
    for component_type, component_config in yaml_data["components"].items():
        component_dict = import_component_helper(
            component_type,
            component_config,
        )
        component_data[component_type + "_name"] = component_dict

    register_stack_helper(stack_name=stack_name, **component_data)



from typing import ClassVar, Union

from ml_metadata.proto import metadata_store_pb2
from tfx.orchestration import metadata

from coalescenceml.metadata_store import BaseMetadataStore


class MySQLMetadataStore(BaseMetadataStore):
    """MySQL backend for CoalescenceML metadata store."""

    host: str
    port: int
    database: str
    username: str
    password: str

    # Class Configuration
    FLAVOR: ClassVar[str] = "mysql"

    def get_tfx_metadata_config(
        self,
    ) -> Union[
        metadata_store_pb2.ConnectionConfig,
        metadata_store_pb2.MetadataStoreClientConfig,
    ]:
        """Return tfx metadata config for mysql metadata store."""
        return metadata.mysql_metadata_connection_config(
            host=self.host,
            port=self.port,
            database=self.database,
            username=self.username,
            password=self.password,
        )



import typing
from typing import Any, Dict, List, Optional, Tuple, Type

import numpy as np
import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from coalescenceml.step import BaseStep
from coalescenceml.step import BaseStepConfig

class TFClassifierConfig(BaseStepConfig):
    layers: typing.List
    input_shape: Tuple[int, ...] = (28,28) # MNIST Size
    num_classes: int = 2
    learning_rate: float = 1e-3
    epochs: int = 10
    batch_size: int = 32


class TFClassifierTrainStep(BaseStep):
    """ """
    def entrypoint(
        self,
        config: TFClassifierConfig,
        x_train: np.ndarray,
        y_train: np.ndarray,
        x_validation: np.ndarray = np.array([]),
        y_validation: np.ndarray = np.array([]),
    ) -> tf.keras.Model:
        # Possibly import optimizer and change its hyperparameters

        model = Sequential()
        model.add(tf.keras.layers.InputLayer(input_shape=config.input_shape))
        model.add(tf.keras.layers.Flatten())

        last_layer = config.num_classes if config.num_classes > 2 else 1
        # Construct layers
        for i, layer_size in enumerate(config.layers):
            model.add(tf.keras.layers.Dense(layer_size, activation='relu'))

        model.add(Dense(last_layer, activation='linear'))

        if last_layer == 1:
            loss = tf.keras.losses.BinaryCrossentropy(from_logits=True)
        else:
            loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)

        model.compile(
            loss=loss,
            optimizer=tf.keras.optimizers.SGD(
                learning_rate=config.learning_rate,
                momentum=0.9,
            ),
            metrics=["accuracy"],
        )

        if len(x_validation) == 0:
            checkpoint_callback = ModelCheckpoint(
                filepath="best_model.hdf5",
                monitor="val_accuracy",
                save_best_only=True,
                save_weights_only=True,
            )
            early_stop = EarlyStopping(
                monitor="val_accuracy",
                mode="max",
                patience=20,
            )
        else:
            checkpoint_callback = ModelCheckpoint(
                filepath="best_model.hdf5",
                monitor="accuracy",
                save_best_only=True,
                save_weights_only=True,
            )
            early_stop = EarlyStopping(
                monitor="accuracy", mode="max", patience=20,
            )

        model.fit(
            x_train,
            y_train,
            batch_size=config.batch_size,
            epochs=config.epochs,
            callbacks=[checkpoint_callback, early_stop],
            validation_data=((x_validation, y_validation) if len(x_validation) > 0 else None),
        )

        model.load_weights("best_model.hdf5")

        return model



from typing import AbstractSet, Callable, Optional, Tuple

from coalescenceml.enums import StackComponentFlavor
from coalescenceml.logger import get_logger
from coalescenceml.stack import Stack
from coalescenceml.stack.exceptions import StackValidationError


logger = get_logger(__name__)


class StackValidator:
    """A `StackValidator` is used to validate a stack configuration.

    Each `StackComponent` can provide a `StackValidator` to make sure it is
    compatible with all components of the stack. The `KubeflowOrchestrator`
    for example will always require the stack to have a container registry
    in order to push the docker images that are required to run a pipeline
    in Kubeflow Pipelines.
    """

    def __init__(
        self,
        required_components: Optional[AbstractSet[StackComponentFlavor]] = None,
        custom_validation_function: Optional[
            Callable[[Stack], Tuple[bool, str]]
        ] = None,
    ):
        """Initializes a `StackValidator` instance.

        Args:
            required_components: Optional set of stack components that must
                exist in the stack.
            custom_validation_function: Optional function that returns whether
                a stack is valid and an err msg if the stack is invalid.
        """
        self._required_components = required_components or set()
        self._custom_validation_function = custom_validation_function

    def validate(self, stack: Stack) -> None:
        """Validates the given stack.

        Checks if the stack contains all the required components and passes
        the custom validation function of the validator.

        Raises:
            StackValidationError: If the stack does not meet all the
                validation criteria.
        """
        missing_components = self._required_components - set(stack.components)
        if missing_components:
            raise StackValidationError(
                f"Missing stack components {missing_components} for "
                f"stack: {stack}"
            )

        if (self._custom_validation_function):
            valid_stack, err_msg = self._custom_validation_function(stack)
            if not valid_stack:
                raise StackValidationError(
                    f"Custom validation function failed to validate "
                    f"stack '{stack.name}': {err_msg}"
                )



from __future__ import annotations

import os
import time
import uuid
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    NoReturn,
    Optional,
    Set,
    Type,
)

from coalescenceml.config.global_config import GlobalConfiguration
from coalescenceml.constants import RUN_NAME_OPTION_KEY
from coalescenceml.enums import StackComponentFlavor
from coalescenceml.io import utils
from coalescenceml.logger import get_logger
from coalescenceml.stack.exceptions import ProvisioningError, StackValidationError
from coalescenceml.utils import readability_utils


if TYPE_CHECKING:
    from coalescenceml.artifact_store import BaseArtifactStore
    from coalescenceml.container_registry import BaseContainerRegistry
    from coalescenceml.experiment_tracker import BaseExperimentTracker
    # from coalescenceml.feature_store import BaseFeatureStore
    from coalescenceml.metadata_store import BaseMetadataStore
    # from coalescenceml.model_deployer import BaseModelDeployer
    from coalescenceml.orchestrator import BaseOrchestrator
    from coalescenceml.pipeline import BasePipeline
    from coalescenceml.pipeline.runtime_configuration import (
        RuntimeConfiguration,
    )

    # from coalescenceml.secrets_manager import BaseSecretsManager
    from coalescenceml.stack import StackComponent
    from coalescenceml.step_operator import BaseStepOperator

logger = get_logger(__name__)


class Stack:
    """CoalescenceML stack class.

    A CoalescenceML stack is a collection of multiple stack components that are
    required to run CoalescenceML pipelines. Some of these components
    (orchestrator, metadata store and artifact store) are required to run
    any kind of pipeline, other components like the container registry are
    only required if other stack components depend on them.
    """

    def __init__(
        self,
        name: str,
        *,
        orchestrator: BaseOrchestrator,
        metadata_store: BaseMetadataStore,
        artifact_store: BaseArtifactStore,
        container_registry: Optional[BaseContainerRegistry] = None,
        # secrets_manager: Optional["BaseSecretsManager"] = None,
        step_operator: Optional[BaseStepOperator] = None,
        # feature_store: Optional["BaseFeatureStore"] = None,
        # model_deployer: Optional["BaseModelDeployer"] = None,
        experiment_tracker: Optional[BaseExperimentTracker] = None,
    ):
        """Initialize and validate a stack instance.

        Args:
            name: Name given to this stack
            orchestrator: Orchestrator instance to be used in stack
            metadata_store: Metadata store instance to be used in stack
            artifact_store: Artifact store instance to be used in stack
            container_registry: Container registry instance to be used in stack
            step_operator: Step operator instance to be used in stack
            experiment_tracker: Experiment tracker instance to be used
        """
        self._name = name
        self._orchestrator = orchestrator
        self._metadata_store = metadata_store
        self._artifact_store = artifact_store
        self._container_registry = container_registry
        self._step_operator = step_operator
        self._secrets_manager = None  # secrets_manager
        self._feature_store = None  # feature_store
        self._model_deployer = None  # model_deployer
        self._experiment_tracker = experiment_tracker

        self.validate()

    @classmethod
    def from_components(
        cls, name: str, components: Dict[StackComponentFlavor, StackComponent]
    ) -> "Stack":
        """Create a stack instance from a dict of stack components.

        Args:
            name: The name of the stack.
            components: The components of the stack.

        Returns:
            A stack instance consisting of the given components.
        """
        from coalescenceml.artifact_store import BaseArtifactStore
        from coalescenceml.container_registry import BaseContainerRegistry
        from coalescenceml.experiment_tracker import BaseExperimentTracker
        # from coalescenceml.feature_store import BaseFeatureStore
        from coalescenceml.metadata_store import BaseMetadataStore
        # from coalescenceml.model_deployer import BaseModelDeployer
        from coalescenceml.orchestrator import BaseOrchestrator
        # from coalescenceml.secrets_manager import BaseSecretsManager
        from coalescenceml.step_operator import BaseStepOperator

        def _raise_type_error(
            component: Optional[StackComponent], expected_class: Type[Any]
        ) -> NoReturn:
            """Raise a TypeError that the component has an unexpected type.

            Args:
                component: StackComponent that has the wrong type
                expected_class: Type of StackComponent that was expected

            Raises:
                TypeError: Always!
            """
            raise TypeError(
                f"Unable to create stack: Wrong stack component type "
                f"`{component.__class__.__name__}` (expected: subclass "
                f"of `{expected_class.__name__}`)"
            )

        orchestrator = components.get(StackComponentFlavor.ORCHESTRATOR)
        if not isinstance(orchestrator, BaseOrchestrator):
            _raise_type_error(orchestrator, BaseOrchestrator)

        metadata_store = components.get(StackComponentFlavor.METADATA_STORE)
        if not isinstance(metadata_store, BaseMetadataStore):
            _raise_type_error(metadata_store, BaseMetadataStore)

        artifact_store = components.get(StackComponentFlavor.ARTIFACT_STORE)
        if not isinstance(artifact_store, BaseArtifactStore):
            _raise_type_error(artifact_store, BaseArtifactStore)

        container_registry = components.get(
            StackComponentFlavor.CONTAINER_REGISTRY
        )
        if container_registry is not None and not isinstance(
            container_registry, BaseContainerRegistry
        ):
            _raise_type_error(container_registry, BaseContainerRegistry)

        experiment_tracker = components.get(
            StackComponentFlavor.EXPERIMENT_TRACKER
        )
        if experiment_tracker is not None and not isinstance(
            experiment_tracker, BaseExperimentTracker
        ):
            _raise_type_error(experiment_tracker, BaseExperimentTracker)

        # secrets_manager = components.get(
        #     StackComponentFlavor.SECRETS_MANAGER
        # )
        # if secrets_manager is not None and not isinstance(
        #     secrets_manager, BaseSecretsManager
        # ):
        #     _raise_type_error(secrets_manager, BaseSecretsManager)

        step_operator = components.get(StackComponentFlavor.STEP_OPERATOR)
        if step_operator is not None and not isinstance(
            step_operator, BaseStepOperator
        ):
            _raise_type_error(step_operator, BaseStepOperator)

        # feature_store = components.get(StackComponentFlavor.FEATURE_STORE)
        # if feature_store is not None and not isinstance(
        #     feature_store, BaseFeatureStore
        # ):
        #     _raise_type_error(feature_store, BaseFeatureStore)

        # model_deployer = components.get(StackComponentFlavor.MODEL_DEPLOYER)
        # if model_deployer is not None and not isinstance(
        #     model_deployer, BaseModelDeployer
        # ):
        #     _raise_type_error(model_deployer, BaseModelDeployer)

        return Stack(
            name=name,
            orchestrator=orchestrator,
            metadata_store=metadata_store,
            artifact_store=artifact_store,
            container_registry=container_registry,
            # secrets_manager=secrets_manager,
            step_operator=step_operator,
            # feature_store=feature_store,
            # model_deployer=model_deployer,
            experiment_tracker=experiment_tracker,
        )

    @classmethod
    def default_local_stack(cls) -> Stack:
        """Create a stack instance which is configured to run locally.

        Returns:
            Default setup for a local stack
        """
        from coalescenceml.artifact_store import LocalArtifactStore
        from coalescenceml.metadata_store import SQLiteMetadataStore
        from coalescenceml.orchestrator import LocalOrchestrator

        orchestrator = LocalOrchestrator(name="default")

        artifact_store_uuid = uuid.uuid4()
        artifact_store_path = os.path.join(
            GlobalConfiguration().config_directory,
            "local_stores",
            str(artifact_store_uuid),
        )
        utils.create_dir_recursive_if_not_exists(artifact_store_path)
        artifact_store = LocalArtifactStore(
            name="default",
            uuid=artifact_store_uuid,
            path=artifact_store_path,
        )

        metadata_store_path = os.path.join(artifact_store_path, "metadata.db")
        metadata_store = SQLiteMetadataStore(
            name="default", uri=metadata_store_path
        )

        return cls(
            name="default",
            orchestrator=orchestrator,
            metadata_store=metadata_store,
            artifact_store=artifact_store,
        )

    @property
    def components(self) -> Dict[StackComponentFlavor, StackComponent]:
        """Fetch all components of the stack.

        Returns:
            All set components of stack
        """
        return {
            component.TYPE: component
            for component in [
                self._orchestrator,
                self._metadata_store,
                self._artifact_store,
                self._container_registry,
                self._secrets_manager,
                self._step_operator,
                self._feature_store,
                self._model_deployer,
                self._experiment_tracker,
            ]
            if component is not None
        }

    @property
    def name(self) -> str:
        """Fetch name of the stack.

        Returns:
            name within stack
        """
        return self._name

    @property
    def orchestrator(self) -> "BaseOrchestrator":
        """Fetch orchestrator of the stack.

        Returns:
            orchestrator within stack
        """
        return self._orchestrator

    @property
    def metadata_store(self) -> "BaseMetadataStore":
        """Fetch metadata store of the stack.

        Returns:
            metadata store within stack
        """
        return self._metadata_store

    @property
    def artifact_store(self) -> "BaseArtifactStore":
        """Fetch artifact store of the stack.

        Returns:
            artifact store within stack
        """
        return self._artifact_store

    @property
    def container_registry(self) -> Optional["BaseContainerRegistry"]:
        """Fetch container registry of the stack.

        Returns:
            container registry within stack
        """
        return self._container_registry

    # @property
    # def secrets_manager(self) -> Optional["BaseSecretsManager"]:
    #     """The secrets manager of the stack."""
    #     return self._secrets_manager

    @property
    def step_operator(self) -> Optional["BaseStepOperator"]:
        """Fetch step operator of the stack.

        Returns:
            step operator within stack
        """
        return self._step_operator

    # @property
    # def feature_store(self) -> Optional["BaseFeatureStore"]:
    #     """The feature store of the stack."""
    #     return self._feature_store

    # @property
    # def model_deployer(self) -> Optional["BaseModelDeployer"]:
    #     """The model deployer of the stack."""
    #     return self._model_deployer

    @property
    def experiment_tracker(self) -> Optional[BaseExperimentTracker]:
        """Fetch experiment tracker of the stack.

        Returns:
            experiment tracker within the stack
        """
        return self._experiment_tracker

    @property
    def runtime_options(self) -> Dict[str, Any]:
        """Runtime options that are available to configure this stack.

        This method combines the available runtime options for all components
        of this stack. See `StackComponent.runtime_options()` for
        more information.

        Returns:
            All configurable runtime options
        """
        runtime_options: Dict[str, Any] = {}
        for component in self.components.values():
            duplicate_runtime_options = (
                runtime_options.keys() & component.runtime_options.keys()
            )
            if duplicate_runtime_options:
                logger.warning(
                    "Found duplicate runtime options %s.",
                    duplicate_runtime_options,
                )

            runtime_options.update(component.runtime_options)

        return runtime_options

    def dict(self) -> Dict[str, Any]:
        """Convert the stack into a dictionary.

        Returns:
            Stack in dictionary form
        """
        component_dict = {
            component_type.value: component.json(sort_keys=True)
            for component_type, component in self.components.items()
        }
        component_dict.update({"name": self.name})
        return component_dict

    def requirements(
        self,
        exclude_components: Optional[AbstractSet[StackComponentFlavor]] = None,
    ) -> Set[str]:
        """Collect set of PyPI requirements for the stack.

        This method combines the requirements of all stack components (except
        the ones specified in `exclude_components`).

        Args:
            exclude_components: Set of component types for which the
                requirements should not be included in the output.

        Returns:
            Set of PyPI requirments for stack
        """
        exclude_components = exclude_components or set()
        requirements = [
            component.requirements
            for component in self.components.values()
            if component.TYPE not in exclude_components
        ]
        return set.union(*requirements) if requirements else set()

    def validate(self) -> None:
        """Check whether the stack configuration is valid.

        To check if a stack configuration is valid, the following criteria must
        be met:
        - all components must support the execution mode (either local or
         remote execution) specified by the orchestrator of the stack
        - the `StackValidator` of each stack component has to validate the
         stack to make sure all the components are compatible with each other
        """
        for component in self.components.values():
            if component.validator:
                component.validator.validate(stack=self)

    def deploy_pipeline(
        self,
        pipeline: BasePipeline,
        runtime_configuration: RuntimeConfiguration,
    ) -> Any:
        """Deploy a pipeline on this stack.

        Args:
            pipeline: The pipeline to deploy.
            runtime_configuration: Contains all the runtime configuration
                options specified for the pipeline run.

        Returns:
            The return value of the call to `orchestrator.run_pipeline(...)`.
        """
        self.validate()

        for component in self.components.values():
            if not component.is_running:
                raise StackValidationError(
                    f"The '{component.name}' {component.TYPE} stack component "
                    f"is not currently running. Please run the following to "
                    f"provision and start the component:\n\n"
                    f"    coml stack up\n"
                )

        for component in self.components.values():
            component.prepare_pipeline_deployment(
                pipeline=pipeline,
                stack=self,
                runtime_configuration=runtime_configuration,
            )

        for component in self.components.values():
            component.prepare_pipeline_run()

        runtime_configuration[
            RUN_NAME_OPTION_KEY
        ] = runtime_configuration.run_name or (
            f"{pipeline.name}-"
            f'{datetime.now().strftime("%d_%h_%y-%H_%M_%S_%f")}'
        )

        logger.info(
            "Using stack `%s` to run pipeline `%s`...",
            self.name,
            pipeline.name,
        )
        start_time = time.time()

        original_cache_boolean = pipeline.enable_cache
        if "enable_cache" in runtime_configuration:
            logger.info(
                "Runtime configuration overwriting the pipeline cache settings"
                " to enable_cache=`%s` for this pipeline run. The default "
                "caching strategy is retained for future pipeline runs.",
                runtime_configuration["enable_cache"],
            )
            pipeline.enable_cache = runtime_configuration.get("enable_cache")

        return_value = self._orchestrator.run_pipeline(
            pipeline, stack=self, runtime_configuration=runtime_configuration
        )

        # Put pipeline level cache policy back to make sure the next runs
        #  default to that policy again in case the runtime configuration
        #  is not set explicitly
        pipeline.enable_cache = original_cache_boolean

        run_duration = time.time() - start_time
        logger.info(
            "Pipeline run `%s` has finished in %s.",
            runtime_configuration.run_name,
            readability_utils.get_human_readable_time(run_duration),
        )

        for component in self.components.values():
            component.cleanup_pipeline_run()

        return return_value

    def prepare_step_run(self) -> None:
        """Prepare running a step."""
        for component in self.components.values():
            component.prepare_step_run()

    def cleanup_step_run(self) -> None:
        """Cleanup resources after step is run."""
        for component in self.components.values():
            component.cleanup_step_run()

    @property
    def is_provisioned(self) -> bool:
        """If the stack provisioned resources to run locally.

        Returns:
            True if stack is provisioned to run locally
        """
        return all(
            component.is_provisioned for component in self.components.values()
        )

    @property
    def is_running(self) -> bool:
        """If the stack is running locally.

        Returns:
            Whether stack is running
        """
        return all(
            component.is_running for component in self.components.values()
        )

    def provision(self) -> None:
        """Provisions resources to run the stack locally."""
        logger.info("Provisioning resources for stack '%s'.", self.name)
        for component in self.components.values():
            if not component.is_provisioned:
                component.provision()
                logger.info("Provisioned resources for %s.", component)

    def deprovision(self) -> None:
        """Deprovisions all local resources of the stack."""
        logger.info("Deprovisioning resources for stack '%s'.", self.name)
        for component in self.components.values():
            if component.is_provisioned:
                try:
                    component.deprovision()
                    logger.info("Deprovisioned resources for %s.", component)
                except NotImplementedError as e:
                    logger.warning(e)

    def resume(self) -> None:
        """Resume the provisioned local resources of the stack.

        Raises:
            ProvisioningError: If any stack component is missing provisioned
                resources.
        """
        logger.info("Resuming provisioned resources for stack %s.", self.name)
        for component in self.components.values():
            if component.is_running:
                # the component is already running, no need to resume anything
                pass
            elif component.is_provisioned:
                component.resume()
                logger.info("Resumed resources for %s.", component)
            else:
                raise ProvisioningError(
                    f"Unable to resume resources for {component}: No "
                    f"resources have been provisioned for this component."
                )

    def suspend(self) -> None:
        """Suspends the provisioned local resources of the stack."""
        logger.info(
            "Suspending provisioned resources for stack '%s'.", self.name
        )
        for component in self.components.values():
            if not component.is_suspended:
                try:
                    component.suspend()
                    logger.info("Suspended resources for %s.", component)
                except NotImplementedError:
                    logger.warning(
                        "Suspending provisioned resources not implemented "
                        "for %s. Continuing without suspending resources...",
                        component,
                    )



from pathlib import Path
from typing import Any, Dict

import yaml

from coalescenceml.io import fileio, utils


def write_yaml(file_path: str, contents: Dict[Any, Any]) -> None:
    """Write contents as YAML format to file_path.
    Args:
        file_path: Path to YAML file.
        contents: Contents of YAML file as dict.
    Raises:
        FileNotFoundError: if directory does not exist.
    """
    if not utils.is_remote(file_path):
        dir_ = str(Path(file_path).parent)
        if not fileio.isdir(dir_):
            raise FileNotFoundError(f"Directory {dir_} does not exist.")
    utils.write_file_contents_as_string(file_path, yaml.dump(contents))


def append_yaml(file_path: str, contents: Dict[Any, Any]) -> None:
    """Append contents to a YAML file at file_path."""
    file_contents = read_yaml(file_path) or {}
    file_contents.update(contents)
    if not utils.is_remote(file_path):
        dir_ = str(Path(file_path).parent)
        if not fileio.isdir(dir_):
            raise FileNotFoundError(f"Directory {dir_} does not exist.")
    utils.write_file_contents_as_string(file_path, yaml.dump(file_contents))


def read_yaml(file_path: str) -> Any:
    """Read YAML on file path and returns contents as dict.
    Args:
        file_path: Path to YAML file.
    Returns:
        Contents of the file in a dict.
    Raises:
        FileNotFoundError: if file does not exist.
    """

    if fileio.exists(file_path):
        contents = utils.read_file_contents_as_string(file_path)
        # TODO: [LOW] consider adding a default empty dict to be returned
        #   instead of None
        return yaml.safe_load(contents)
    else:
        raise FileNotFoundError(f"{file_path} does not exist.")


def is_yaml(file_path: str) -> bool:
    """Returns True if file_path is YAML, else False
    Args:
        file_path: Path to YAML file.
    Returns:
        True if is yaml, else False.
    """
    if file_path.endswith("yaml") or file_path.endswith("yml"):
        return True
    return False



from typing import Any, ClassVar, Tuple, Type

from coalescenceml.artifacts.base_artifact import BaseArtifact


class BaseProducer(object):
    """BaseProducer to produce artifact data."""

    ARTIFACT_TYPES: ClassVar[Tuple[Type[BaseArtifact], ...]] = ()
    TYPES: ClassVar[Tuple[Type[Any], ...]] = ()

    def __init__(self, artifact: BaseArtifact):
        self.artifact = artifact

    def can_handle_type(self, data_type: Type[Any]) -> bool:
        """can_handle_type determines whether the producer can r/w a certain type.

        Args:
            data_type: the type to check for

        Returns:
            can_handle: Whether the type can be r/w
        """
        return any(issubclass(data_type, t) for t in self.TYPES)

    def handle_input(self, data: Type[Any]) -> Any:
        """ """
        if not self.can_handle_type(data):
            raise TypeError(f"Unable to handle type {data}.")

    def handle_return(self, data: Any) -> None:
        """ """
        data_type = type(data)
        if not self.can_handle_type(data_type):
            raise TypeError(f"Unable to handle type {data}.")



import logging
from enum import Enum
from typing import List


class DictEnum(str, Enum):
    """DictEnum is a base enum type of key values pairs like a dict.

    ..note: This is a subclass of str as well for nice properties with respect
        to logging.
    """
    def __str__(self)->str:
        return self.value

    @classmethod
    def names(cls) -> List[str]:
        """Get all enum names as a list of strings.

        Returns:
            List of names of all enums
        """
        return [c.name for c in cls]

    @classmethod
    def values(cls) -> List[str]:
        """Get all enum values as a list of strings.

        Returns:
            List of vlues of all enums
        """
        return [c.value for c in cls]


class LoggingLevels(Enum):
    """LoggingLevels is an enum for logging levels."""

    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARN = logging.WARN
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class ExecutionStatus(DictEnum):
    """Reprent the current status of a step or pipeline run."""

    FAILED = "failed"
    RUNNING = "running"
    CACHED = "cached"
    COMPLETED = "completed"


class StackComponentFlavor(DictEnum):
    """All possible types a `StackComponent` can have."""

    ORCHESTRATOR = "orchestrator"
    METADATA_STORE = "metadata_store"
    ARTIFACT_STORE = "artifact_store"
    CONTAINER_REGISTRY = "container_registry"
    STEP_OPERATOR = "step_operator"  # TODO
    EXPERIMENT_TRACKER = "experiment_tracker"
    FEATURE_STORE = "feature_store"  # TODO
    SECRETS_MANAGER = "secrets_manager"  # TODO
    MODEL_DEPLOYER = "model_deployer"  # TODO


class DirectoryStoreFlavor(DictEnum):
    """Directory Store Backend Types."""

    LOCAL = "local"
    SQL = "sql"


class MetadataContextFlavor(DictEnum):
    """All possible types that contexts can have within pipeline nodes."""

    STACK = "stack"
    PIPELINE_REQUIREMENTS = "pipeline_requirements"



import re
from typing import ClassVar

from coalescenceml.enums import StackComponentFlavor
from coalescenceml.stack import StackComponent


class BaseContainerRegistry(StackComponent):
    """Base class for all CoalescenceML container registries.
    Attributes:
        uri: The URI of the container registry.
    """

    uri: str

    # Class Configuration
    TYPE: ClassVar[
        StackComponentFlavor
    ] = StackComponentFlavor.CONTAINER_REGISTRY
    FLAVOR: ClassVar[str] = "default"

    @property
    def is_local(self) -> bool:
        return re.fullmatch(r"localhost:[0-9]{4,5}", self.uri)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, cast

from coalescenceml.constants import RUN_NAME_OPTION_KEY
from coalescenceml.logger import get_logger


if TYPE_CHECKING:
    from coalescenceml.pipeline import Schedule

logger = get_logger(__name__)
SCHEDULE_OPTION_KEY = "schedule"


class RuntimeConfiguration(Dict[str, Any]):
    """RuntimeConfiguration store dynamic options for a pipeline run.
    Use `stack.runtime_options()` to get all available runtime options for the
    components of a specific CoalescenceML stack.
    This class is a `dict` subclass, so getting/setting runtime options is done
    using `some_value = runtime_configuration["some_key"]` and
    `runtime_configuration["some_key"] = 1`.
    """

    def __init__(
        self,
        *,
        run_name: Optional[str] = None,
        schedule: Optional["Schedule"] = None,
        **runtime_options: Any,
    ):
        """Initializes a RuntimeConfiguration object.
        Args:
            run_name: Optional name of the pipeline run.
            schedule: Optional schedule of the pipeline run.
            **runtime_options: Additional runtime options.
        """
        runtime_options[RUN_NAME_OPTION_KEY] = run_name
        runtime_options[SCHEDULE_OPTION_KEY] = schedule
        super().__init__(runtime_options)

    @property
    def run_name(self) -> Optional[str]:
        """Name of the pipeline run."""
        return cast(Optional[str], self[RUN_NAME_OPTION_KEY])

    @property
    def schedule(self) -> Optional[Schedule]:
        """Schedule of the pipeline run."""
        from coalescenceml.pipeline import Schedule

        return cast(Optional[Schedule], self[SCHEDULE_OPTION_KEY])



from coalescenceml.integrations.tensorflow.producers.keras_model_producer import (
    KerasModelProducer,
)
from coalescenceml.integrations.tensorflow.producers.tf_dataset_producer import (
    TensorflowDatasetProducer,
)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['coalescenceml',
 'coalescenceml.artifact_store',
 'coalescenceml.artifacts',
 'coalescenceml.cli',
 'coalescenceml.config',
 'coalescenceml.container_registry',
 'coalescenceml.experiment_tracker',
 'coalescenceml.integrations',
 'coalescenceml.integrations.aws_s3',
 'coalescenceml.integrations.aws_s3.artifact_store',
 'coalescenceml.integrations.azure_datalake',
 'coalescenceml.integrations.azure_datalake.artifact_store',
 'coalescenceml.integrations.mlflow',
 'coalescenceml.integrations.mlflow.experiment_tracker',
 'coalescenceml.integrations.sklearn',
 'coalescenceml.integrations.sklearn.producers',
 'coalescenceml.integrations.sklearn.step',
 'coalescenceml.integrations.statsmodels',
 'coalescenceml.integrations.statsmodels.producers',
 'coalescenceml.integrations.tensorflow',
 'coalescenceml.integrations.tensorflow.producers',
 'coalescenceml.integrations.tensorflow.step',
 'coalescenceml.integrations.wandb',
 'coalescenceml.integrations.wandb.experiment_tracker',
 'coalescenceml.integrations.xgboost',
 'coalescenceml.integrations.xgboost.producers',
 'coalescenceml.io',
 'coalescenceml.metadata_store',
 'coalescenceml.orchestrator',
 'coalescenceml.pipeline',
 'coalescenceml.post_execution',
 'coalescenceml.producers',
 'coalescenceml.stack',
 'coalescenceml.stack_store',
 'coalescenceml.stack_store.model',
 'coalescenceml.step',
 'coalescenceml.step_operator',
 'coalescenceml.utils']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'distro>=1.7.0,<2.0.0',
 'ml-pipelines-sdk>=1.6.1,<2.0.0',
 'numpy>=1.21.6,<2.0.0',
 'pandas>=1.2.0,<2.0.0',
 'pyarrow>=7.0.0,<8.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pyyaml>=5.4.1,<6.0.0',
 'rich[jupyter]>=12.0.0,<13.0.0',
 'semver>=2.13.0,<3.0.0',
 'sqlmodel>=0.0.6,<0.1.0']

entry_points = \
{'console_scripts': ['coml = coalescenceml.cli.cli:cli']}

setup_kwargs = {
    'name': 'coalescenceml',
    'version': '0.5.1',
    'description': 'An open-source MLOps framework to develop industry-grade production ML pipelines coalescing the MLOps stack under one umbrella.',
    'long_description': '# CoalescenceML\n\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/coalescenceml)](https://pypi.org/project/coalescenceml/)\n[![PyPI Status](https://pepy.tech/badge/coalescenceml)](https://pepy.tech/project/coalescenceml)\n[![codecov](https://codecov.io/gh/bayoumi17m/CoalescenceML/branch/main/graph/badge.svg?token=7QNV6GV4B3)](https://codecov.io/gh/bayoumi17m/CoalescenceML)\n![GitHub](https://img.shields.io/github/license/bayoumi17m/CoalescenceML)\n[![Interrogate](docs/interrogate.svg)](https://interrogate.readthedocs.io/en/latest/)\n![Main Workflow Tests](https://github.com/bayoumi17m/CoalescenceML/actions/workflows/main.yml/badge.svg)\n\n# What is Coalescence ML?\n\n# Why use Coalescence ML?\n\n# Learn more about Coalescence ML\n\n## Learn more about MLOps\n\n# Features\n\n# Getting Started\n\n## Install Coalescence ML\n\n## Quickstart\n\n',
    'author': 'Magd Bayoumi, Rafael Chaves, Elva Gao, Sanjali Jha, Iris Li, Jerry Sun, Emily Wang',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.9',
}


setup(**setup_kwargs)



from typing import Any, Dict, List, Tuple


class ConfigKeys:
    """Class to validate dictionary configurations."""

    @classmethod
    def get_keys(cls) -> Tuple[List[str], List[str]]:
        """Gets all the required and optional config keys for this class.
        Returns:
            A tuple (required, optional) which are lists of the
            required/optional keys for this class.
        """
        keys = {
            key: value
            for key, value in cls.__dict__.items()
            if not isinstance(value, classmethod)
            and not isinstance(value, staticmethod)
            and not callable(value)
            and not key.startswith("__")
        }

        required = [v for k, v in keys.items() if not k.endswith("_")]
        optional = [v for k, v in keys.items() if k.endswith("_")]

        return required, optional

    @classmethod
    def key_check(cls, config: Dict[str, Any]) -> None:
        """Checks whether a configuration dict contains all required keys
        and no unknown keys.
        Args:
            config: The configuration dict to verify.
        Raises:
            TypeError: If no config dictionary is passed.
            ValueError: If required keys are missing or unknown keys are found.
        """
        if not isinstance(config, dict):
            raise TypeError(f"Please specify a dict for {cls.__name__}")

        # Required and optional keys for the config dict
        required, optional = cls.get_keys()

        # Check for missing keys
        missing_keys = [k for k in required if k not in config.keys()]
        if missing_keys:
            raise ValueError(f"Missing key(s) {missing_keys} in {cls.__name__}")

        # Check for unknown keys
        unknown_keys = [
            k for k in config.keys() if k not in required and k not in optional
        ]
        if unknown_keys:
            raise ValueError(
                f"Unknown key(s) {unknown_keys} in {cls.__name__}. "
                f"Required keys: {required}, optional keys: {optional}."
            )


class PipelineConfigurationKeys(ConfigKeys):
    """Keys for a pipeline configuration dict."""

    NAME = "name"
    STEPS = "steps"


class StepConfigurationKeys(ConfigKeys):
    """Keys for a step configuration dict."""

    SOURCE_ = "source"
    PARAMETERS_ = "parameters"
    MATERIALIZERS_ = "materializers"



import textwrap
from abc import ABC
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from pydantic import root_validator
from tfx.dsl.io.fileio import NotFoundError

from coalescenceml.artifact_store.exceptions import ArtifactStoreInterfaceError
from coalescenceml.enums import StackComponentFlavor
from coalescenceml.stack import StackComponent


PathType = Union[bytes, str]


# Bullshit hack b/c of TFX....
def _catch_not_found_error(_func: Callable[..., Any]) -> Callable[..., Any]:
    """Utility decorator used for catching a `FileNotFoundError` and
    converting it to a `NotFoundError` in order to deal with the TFX exception
    handling."""

    def inner_function(*args: Any, **kwargs: Any) -> Any:
        try:
            return _func(*args, **kwargs)
        except FileNotFoundError as e:
            raise NotFoundError() from e

    return inner_function


class BaseArtifactStore(StackComponent, ABC):
    """Base class for all CoalescenceML artifact stores.
    Attributes:
        path: The root path of the artifact store.
    """

    path: str

    # Class Configuration
    TYPE: ClassVar[StackComponentFlavor] = StackComponentFlavor.ARTIFACT_STORE
    FLAVOR: ClassVar[str]
    SUPPORTED_SCHEMES: ClassVar[Set[str]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initiate the Pydantic object and register the corresponding
        filesystem."""
        super(BaseArtifactStore, self).__init__(*args, **kwargs)
        self._register()

    @staticmethod
    def open(name: PathType, mode: str = "r") -> Any:
        """Open a file at the given path."""
        raise NotImplementedError()

    @staticmethod
    def copyfile(src: PathType, dst: PathType, overwrite: bool = False) -> None:
        """Copy a file from the source to the destination."""
        raise NotImplementedError()

    @staticmethod
    def exists(path: PathType) -> bool:
        """Returns `True` if the given path exists."""
        raise NotImplementedError()

    @staticmethod
    def glob(pattern: PathType) -> List[PathType]:
        """Return the paths that match a glob pattern."""
        raise NotImplementedError()

    @staticmethod
    def isdir(path: PathType) -> bool:
        """Returns whether the given path points to a directory."""
        raise NotImplementedError()

    @staticmethod
    def listdir(path: PathType) -> List[PathType]:
        """Returns a list of files under a given directory in the filesystem."""
        raise NotImplementedError()

    @staticmethod
    def makedirs(path: PathType) -> None:
        """Make a directory at the given path, recursively creating parents."""
        raise NotImplementedError()

    @staticmethod
    def mkdir(path: PathType) -> None:
        """Make a directory at the given path; parent directory must exist."""
        raise NotImplementedError()

    @staticmethod
    def remove(path: PathType) -> None:
        """Remove the file at the given path. Dangerous operation."""
        raise NotImplementedError()

    @staticmethod
    def rename(src: PathType, dst: PathType, overwrite: bool = False) -> None:
        """Rename source file to destination file."""
        raise NotImplementedError()

    @staticmethod
    def rmtree(path: PathType) -> None:
        """Deletes dir recursively. Dangerous operation."""
        raise NotImplementedError()

    @staticmethod
    def stat(path: PathType) -> Any:
        """Return the stat descriptor for a given file path."""
        raise NotImplementedError()

    @staticmethod
    def walk(
        top: PathType,
        topdown: bool = True,
        onerror: Optional[Callable[..., None]] = None,
    ) -> Iterable[Tuple[PathType, List[PathType], List[PathType]]]:
        """Return an iterator that walks the contents of the given directory."""
        raise NotImplementedError()

    @root_validator
    def _ensure_artifact_store(cls, values: Dict[str, Any]) -> Any:
        """Validator function for the Artifact Stores. Checks whether
        supported schemes are defined and the given path is supported"""
        try:
            getattr(cls, "SUPPORTED_SCHEMES")
        except AttributeError:
            raise ArtifactStoreInterfaceError(
                textwrap.dedent(
                    """
                    When you are working with any classes which subclass from
                    coalescenceml.artifact_store.BaseArtifactStore please make sure that your class
                    has a ClassVar named `SUPPORTED_SCHEMES` which should hold a set of
                    supported file schemes such as {"s3://"} or {"gcs://"}.
                    Example:
                    class S3ArtifactStore(BaseArtifactStore):
                        ...
                        # Class Variables
                        ...
                        SUPPORTED_SCHEMES: ClassVar[Set[str]] = {"s3://"}
                        ...
                    """
                )
            )
        if not any(values["path"].startswith(i) for i in cls.SUPPORTED_SCHEMES):
            raise ArtifactStoreInterfaceError(
                textwrap.dedent(
                    f"""
                    The path: "{values["path"]}" you defined for your artifact
                    store is not supported by the implementation of
                    {cls.schema()["title"]}, because it does not start with
                    one of its supported schemes: {cls.SUPPORTED_SCHEMES}.
                    """
                )
            )

        return values

    def _register(self, priority: int = 5) -> None:
        """Create and register a filesystem within the TFX registry"""
        from tfx.dsl.io.filesystem import Filesystem
        from tfx.dsl.io.filesystem_registry import DEFAULT_FILESYSTEM_REGISTRY

        filesystem_class = type(
            self.__class__.__name__,
            (Filesystem,),
            {
                "SUPPORTED_SCHEMES": self.SUPPORTED_SCHEMES,
                "open": staticmethod(_catch_not_found_error(self.open)),
                "copy": staticmethod(_catch_not_found_error(self.copyfile)),
                "exists": staticmethod(self.exists),
                "glob": staticmethod(self.glob),
                "isdir": staticmethod(self.isdir),
                "listdir": staticmethod(_catch_not_found_error(self.listdir)),
                "makedirs": staticmethod(self.makedirs),
                "mkdir": staticmethod(_catch_not_found_error(self.mkdir)),
                "remove": staticmethod(_catch_not_found_error(self.remove)),
                "rename": staticmethod(_catch_not_found_error(self.rename)),
                "rmtree": staticmethod(_catch_not_found_error(self.rmtree)),
                "stat": staticmethod(_catch_not_found_error(self.stat)),
                "walk": staticmethod(_catch_not_found_error(self.walk)),
            },
        )

        DEFAULT_FILESYSTEM_REGISTRY.register(
            filesystem_class, priority=priority
        )



import importlib
import logging
import sys
from typing import Dict, Type, cast

import click
from tfx.dsl.components.base.base_executor import BaseExecutor
from tfx.orchestration.portable.data_types import ExecutionInfo
from tfx.orchestration.portable.python_executor_operator import (
    run_with_executor,
)
from tfx.proto.orchestration.execution_invocation_pb2 import ExecutionInvocation

from coalescenceml import constants
from coalescenceml.artifacts.base_artifact import BaseArtifact
from coalescenceml.artifacts.type_registry import type_registry
from coalescenceml.integrations.registry import integration_registry
from coalescenceml.io import fileio
from coalescenceml.step import BaseStep
from coalescenceml.step.utils import _FunctionExecutor, generate_component_class
from coalescenceml.utils import json_utils, source_utils


def create_executor_class(
    step_source_path: str,
    input_artifact_type_mapping: Dict[str, str],
) -> Type[_FunctionExecutor]:
    """Creates an executor class for a given step.
    Args:
        step_source_path: Import path of the step to run.
        input_artifact_type_mapping: A dictionary mapping input names to
            a string representation of their artifact classes.
    """
    step_class = cast(
        Type[BaseStep], source_utils.load_source_path_class(step_source_path)
    )
    step_instance = step_class()

    producers = step_instance.get_producers(ensure_complete=True)

    # We don't publish anything to the metadata store inside this environment,
    # so the specific artifact classes don't matter
    input_spec = {}
    for key, value in step_class.INPUT_SIGNATURE.items():
        input_spec[key] = BaseArtifact

    output_spec = {}
    for key, value in step_class.OUTPUT_SIGNATURE.items():
        output_spec[key] = type_registry.get_artifact_type(value)[0]

    execution_parameters = {
        **step_instance.PARAM_SPEC,
        **step_instance._internal_execution_parameters,
    }

    component_class = generate_component_class(
        step_name=step_instance.name,
        step_module=step_class.__module__,
        input_spec=input_spec,
        output_spec=output_spec,
        execution_parameter_names=set(execution_parameters),
        step_function=step_instance.entrypoint,
        producers=producers,
    )

    return cast(
        Type[_FunctionExecutor], component_class.EXECUTOR_SPEC.executor_class
    )


def load_execution_info(execution_info_path: str) -> ExecutionInfo:
    """Loads the execution info from the given path."""
    with fileio.open(execution_info_path, "rb") as f:
        execution_info_proto = ExecutionInvocation.FromString(f.read())

    return ExecutionInfo.from_proto(execution_info_proto)


def configure_executor(
    executor_class: Type[BaseExecutor], execution_info: ExecutionInfo
) -> BaseExecutor:
    """Creates and configures an executor instance.
    Args:
        executor_class: The class of the executor instance.
        execution_info: Execution info for the executor.
    Returns:
        A configured executor instance.
    """
    context = BaseExecutor.Context(
        tmp_dir=execution_info.tmp_dir,
        unique_id=str(execution_info.execution_id),
        executor_output_uri=execution_info.execution_output_uri,
        stateful_working_dir=execution_info.stateful_working_dir,
        pipeline_node=execution_info.pipeline_node,
        pipeline_info=execution_info.pipeline_info,
        pipeline_run_id=execution_info.pipeline_run_id,
    )

    return executor_class(context=context)


@click.command()
@click.option("--main_module", required=True, type=str)
@click.option("--step_source_path", required=True, type=str)
@click.option("--execution_info_path", required=True, type=str)
@click.option("--input_artifact_types_path", required=True, type=str)
def main(
    main_module: str,
    step_source_path: str,
    execution_info_path: str,
    input_artifact_types_path: str,
) -> None:
    """Runs a single CoalescenceML step."""
    # prevent running entire pipeline in user code if they would run at import
    # time (e.g. not wrapped in a function or __name__=="__main__" check)
    constants.SHOULD_PREVENT_PIPELINE_EXECUTION = True

    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    logging.getLogger().setLevel(logging.INFO)

    # activate integrations and import the user main module to register all
    # producers and stack components
    integration_registry.activate_integrations()
    importlib.import_module(main_module)

    input_artifact_type_mapping = json_utils.read_json(
        input_artifact_types_path
    )
    executor_class = create_executor_class(
        step_source_path=step_source_path,
        input_artifact_type_mapping=input_artifact_type_mapping,
    )

    execution_info = load_execution_info(execution_info_path)
    executor = configure_executor(executor_class, execution_info=execution_info)
    run_with_executor(execution_info=execution_info, executor=executor)



"""
## Step Operator

While an orchestrator defines how and where your entire pipeline runs, a step 
operator defines how and where an individual step runs. An example could be
if one step within a pipeline should run on a separate environment equipped
with a GPU (like a trainer step).
"""
from coalescenceml.step_operator.base_step_operator import BaseStepOperator


__all__ = ["BaseStepOperator"]



from typing import Any, Dict

from ml_metadata.proto import metadata_store_pb2
from tfx.types.artifact import Artifact, Property, PropertyType

from coalescenceml.artifacts.constants import (
    DATATYPE_PROPERTY_KEY,
    PRODUCER_PROPERTY_KEY,
)


DATATYPE_PROPERTY = Property(type=PropertyType.STRING)
PRODUCER_PROPERTY = Property(type=PropertyType.STRING)


class BaseArtifact(Artifact):
    """ """

    TYPE_NAME: str = "BaseArtifact"
    PROPERTIES: Dict[str, Property] = {
        DATATYPE_PROPERTY_KEY: DATATYPE_PROPERTY,
        PRODUCER_PROPERTY_KEY: PRODUCER_PROPERTY,
    }
    MLMD_TYPE: Any = None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """"""
        self.validate_and_set_type()
        super(BaseArtifact, self).__init__(*args, **kwargs)

    @classmethod
    def validate_and_set_type(cls) -> None:
        """Validate artifact and set type"""
        type_name = cls.TYPE_NAME
        if not isinstance(type_name, str):
            raise ValueError(
                f"The subclass {cls} must overrise TYPE_NAME attribute with a string type name (got {type_name} instead)"
            )

        # Create ml metadata artifact type
        mlmd_artifact_type = metadata_store_pb2.ArtifactType()
        mlmd_artifact_type.name = type_name  # store the name

        # Perform validation on properties
        if cls.PROPERTIES:
            if not isinstance(cls.PROPERTIES, dict):
                raise ValueError(f"The subclass {cls}.PROPERTIES is not a dict")

            for key, value in cls.PROPERTIES.items():
                if not (isinstance(key, str) and isinstance(value, Property)):
                    raise ValueError(
                        f"The subclass {cls}.PROPERTIES dictionary must have keys of type string and values of type tfx.types.artifact.Property"
                    )

            # Finally populate ML metadata properties
            for key, value in cls.PROPERTIES.items():
                mlmd_artifact_type.properties[key] = value.mlmd_type()
        else:
            raise ValueError("Empty properties dictionary!")

        cls.MLMD_ARTIFACT_TYPE = mlmd_artifact_type



# import fnmatch
# import os
# from pathlib import Path
# from typing import Any, Callable, Iterable, List, Optional, Tuple, Type

# from tfx.dsl.io.filesystem import Filesystem, PathType

# from coalescenceml.constants import REMOTE_FS_PREFIX
# from coalescenceml.io.fileio_registry import fileio_registry
# from coalescenceml.logger import get_logger


# logger = get_logger(__name__)


# def pathtype_to_str(path: PathType) -> str:
#     """Converts PathType to a str.

#     Args:
#         path: location of file.

#     Returns:
#         location of file as a string.
#     """
#     if isinstance(path, str):
#         return path
#     else:
#         return path.decode("utf-8")


# def get_filesystem(path: PathType) -> Type[Filesystem]:
#     """Returns an abstract filesystem class for a given path from the registry.

#     Args:
#         path: path to find associated filesystem for

#     Returns:
#         Filesystem associated with path
#     """
#     return fileio_registry.get_filesystem_for_path(path)


# def open(path: PathType, mode: str = "r") -> Any: # Fix this to be some type of stream.
#     """Open a file at a given path.

#     Args:
#         path: location of file to open.
#         mode: How to open the file [r, w, a]

#     Returns:
#         file stream to operate on.
#     """
#     return get_filesystem(path).open(path, mode=mode)


# def copy(src: PathType, dst: PathType, overwrite: bool = False) -> None:
#     """Copy a file from source to destination.

#     Args:
#         src: source location to grab file.
#         dst: destination location to store file.
#         overwrite: whether to overwrite file if it exists.

#     Raises:
#         FileExistsError: if overwrite is false and file exists in destination
#     """
#     src_fs = get_filesystem(src)
#     tgt_fs = get_filesystem(tgt)

#     if src_fs is dst_fs:
#         src_fs.copy(src, dst, overwrite=overwrite)
#     else:
#         if file_exists(dst) and not overwrite:
#             raise FileExistsError(f"File {pathtype_to_str} in destination already exists and overwrite is false.")
#         with open(src, mode="rb") as fp:
#             contents = fp.read()
#         with open(dst, mode="wb") as fp:
#             fp.write(contents)


# def file_exists(path: PathType) -> bool:
#     """Return True if the given path exists.

#     Args:
#         path: location of file

#     Returns:
#         whether file exists at location.
#     """
#     return get_filesystem(path).exists(path)


# def remove(path: PathType) -> None:
#     """Remove the file at the given path.

#     Args:
#         path: location of file

#     Warning:
#         Dangerous operation! Do at your own peril...

#     Raises:
#         FileNotFoundError: path does not exist
#     """
#     if not file_exists(path):
#         raise FileNotFoundError(f"{pathtype_to_str(path)} does not exist.")
#     get_filesystem(path).remove(path)


# def is_dir(path: PathType) -> bool:
#     """Returns whether path is a directory.

#     Args:
#         path: location of file

#     Returns:
#         True is the path is a directory, otherwise False
#     """
#     return get_filesystem(path).isdir(path)


# def glob(pattern: PathType) -> List[PathType]:
#     """Return paths that match the pattern.

#     Args:
#         patten: expression of paths with wildcards

#     Returns:
#         List of paths that match the pattern
#     """
#     return get_filesystem(path).glob(path)


# def list_dir(dir_path: str, only_file_names: bool = False) -> List[str]:
#     """Returns a list of files in dir.

#     Args:
#         dir_path: Path in filesystem.
#         only_file_names: Returns only file names if True.

#     Returns:
#         List of paths in dir
#     """
#     try:
#         return [
#             pathtype_to_str(f)
#             if only_file_names
#             else os.path.join(dir_path, pathtype_to_str(f))
#             for f in get_filesystem(dir_path).listdir(dir_path)
#         ]
#     except IOError:
#         raise ValueError(f"Dir {dir_path} not found.")


# def make_dirs(path: PathType) -> None:
#     """Make a directory at path, creating parents.

#     Args:
#         path: Path in filesystem to make
#     """
#     get_filesystem(path).makedirs(path)


# def mkdir(path: PathType) -> None:
#     """Make a directory at path, parent directory must exist.

#     Args:
#         path: Path in filesystem to make
#     """
#     get_filesystem(path).mkdir(path)


# def rename(src: PathType, dst: PathType, overwrite: bool = True) -> None:
#     """Rename source file to destination file.

#     Args:
#         src: Path of file to rename.
#         dst: Path of file to rename to.
#         overwrite: If file exists at destination, then it will be overwritten if this is True and raise a FileExistsError otherwise

#     Raises:
#         FileExistsError: If file already exists at the destination and overwrite is set to False.
#         ValueError: If src and dst use different filesystems plugins
#     """
#     src_fs = get_filesystem(src)
#     dst_fs = get_filesystem(dst)
#     if src_fs is dst_fs:
#         src_fs.rename(src, dst, overwrite=overwrite)
#     else:
#         raise ValueError(f"Renaming from different plugins is not supported.")


# def rm_dir(dir_path: str) -> None:
#     """Deletes dir recursively.

#     Args:
#         dir_path: Dir to delete

#     Raises:
#         ValueError: If the path is not pointing to a directory

#     Warnings:
#         Dangerous operation....proceed with causion.
#     """
#     if not is_dir(dir_path):
#         raise ValueError(f"Path '{dir_path}' is not a directory.")

#     get_filesystem(dir_path).rmtree(dir_path)


# def stat(path: PathType) -> Any:
#     """Return stat description for given path.

#     Args:
#         path: Path in filesystem.

#     Returns:
#         stat description of file path
#     """
#     get_filesystem(path).stat(path)


# def walk(root: PathType, topdown: bool = True, onerror: Optional[Callable[..., None]] = None,) -> Iterable[Tuple[PathType, List[PathType], List[PathType]]]:
#     """Return iterator that walks contents of directory.

#     Args:
#         root: Path of directory to walk.
#         topdown: whether to walk directories topdown or bottom-up.
#         onerror: Callable that is run if error occurs.

#     Returns:
#         Iterable of tuples which contain: Path of the current directory path, a list of directories inside the current directory, and a list of files inside the current directory.
#     """
#     return get_filesystem(root).walk(root, topdown=topdown, onerror=onerror)


# def grep_files(dir_path: PathType, pattern: str) -> Iterable[str]:
#     """Find files in directory that match a pattern.

#     Args:
#         dir_path: Path to directory.
#         pattern: file path pattern like *.png

#     Yields:
#         All matching filenames
#     """
#     for root, dirs, files, in walk(dir_path):
#         for basename in files:
#             if fnmatch.fnmatch(convert_to_str(basename), pattern):
#                 filename = os.path.join(pathtype_to_str(root), pathtype_to_str(basename))
#                 yield filename


# def is_remote(path: str) -> bool:
#     """Returns True if path is remote.

#     Args:
#         path: file path in a filesystem as a string

#     Returns:
#         True if remote path else False
#     """
#     return any(path.startswith(prefix) for prefix in REMOTE_FS_PREFIX)


# def create_dir_if_not_exists(dir_path: str) -> None:
#     """Creates directory if it does not exist.

#     Args:
#         dir_path: Local path in filesystem.
#     """
#     if not is_dir(dir_path):
#         make_dirs(dir_path)


# def create_dir_recursive_if_not_exists(dir_path: str) -> None:
#     """Creates directory recursively if it does not exist.

#     Args:
#         dir_path: Local path in filesystem.
#     """
#     if not is_dir(dir_path):
#         make_dirs(dir_path)


# def resolve_relative_path(path: str) -> str:
#     """Takes a relative path and conbverts into an absolute path.

#     Args:
#         path: Local path in filesystem.

#     Returns:
#         Resolved path.
#     """
#     if is_remote(path):
#         return path
#     else:
#         return str(Path(path).resolve())

# def move(src: str, dst: str, overwrite: bool = False) -> None:
#     """Moves dir or file from source to destination.

#     Args:
#         src: Local path to copy from.
#         dst: Local path to copy to.
#         overwrite: boolean, if false, then throws an error before overwriting.
#     """
#     rename(src, dst, overwrite)


# def copy_dir(src_dir: str, dst_dir, overwrite: bool=False) -> None:
#     """Copies dir from src to dst.

#     Args:
#         src_dir: Path to copy folder from.
#         dst_dir: Path to copy folder to.
#         overwrite: whether to overwrite existing dst. If false, throws an error before overwriting.
#     """
#     for src_file in list_dir(src_dir):
#         src_file_path = Path(src_file)
#         dst_name = os.path.join(dst_dir, src_file_path.name)
#         if is_dir(src_file):
#             copy_dir(src_file, dst_name, overwrite)
#         else:
#             create_dir_recursive_if_not_exists(
#                 str(Path(dst_name).parent)
#             )
#             copy(str(src_file_path), str(dst_name), overwrite)


# def get_parent(dir_path: str) -> str:
#     """Get parent of dir.

#     Args:
#         dir_path: Path to directory.

#     Returns:
#         Parent of the dir as a string.
#     """
#     return Path(dir_path).parent.stem

from tfx.dsl.io.fileio import (  # noqa
    copy,
    exists,
    glob,
    isdir,
    listdir,
    makedirs,
    mkdir,
    open,
    remove,
    rename,
    rmtree,
    stat,
    walk,
)


__all__ = [
    "copy",
    "exists",
    "glob",
    "isdir",
    "listdir",
    "makedirs",
    "mkdir",
    "open",
    "remove",
    "rename",
    "rmtree",
    "stat",
    "walk",
]



from collections import defaultdict
from typing import ClassVar, DefaultDict, Dict, Type, TypeVar

from coalescenceml.artifact_store import LocalArtifactStore
from coalescenceml.container_registry import BaseContainerRegistry
from coalescenceml.enums import StackComponentFlavor
from coalescenceml.logger import get_logger
from coalescenceml.metadata_store import MySQLMetadataStore, SQLiteMetadataStore
from coalescenceml.orchestrator import LocalOrchestrator
from coalescenceml.stack import StackComponent


logger = get_logger(__name__)


class StackComponentClassRegistry:
    """Registry for stack component classes.
    All stack component classes must be registered here, so they can be
    instantiated from the component type and flavor specified inside the
    CoalescenceML directory configuration.
    """

    component_classes: ClassVar[
        DefaultDict[StackComponentFlavor, Dict[str, Type[StackComponent]]]
    ] = defaultdict(dict)

    @classmethod
    def register_class(
        cls,
        component_class: Type[StackComponent],
    ) -> None:
        """Registers a stack component class."""
        component_flavor = component_class.FLAVOR
        flavors = cls.component_classes[component_class.TYPE]
        if component_flavor in flavors:
            logger.warning(
                "Overwriting previously registered stack component class `%s` "
                "for type '%s' and flavor '%s'.",
                flavors[component_flavor].__class__.__name__,
                component_class.TYPE.value,
                component_class.FLAVOR,
            )

        flavors[component_flavor] = component_class
        logger.debug(
            "Registered stack component class for type '%s' and flavor '%s'.",
            component_class.TYPE.value,
            component_flavor,
        )

    @classmethod
    def get_class(
        cls,
        component_type: StackComponentFlavor,
        component_flavor: str,
    ) -> Type[StackComponent]:
        """Returns the stack component class for the given type and flavor.

        Args:
            component_type: The type of the component class to return.
            component_flavor: The flavor of the component class to return.
        Raises:
            KeyError: If no component class is registered for the given type
                and flavor.
        """

        available_flavors = cls.component_classes[component_type]
        try:
            return available_flavors[component_flavor]
        except KeyError:
            # The stack component might be part of an integration
            # -> Activate the integrations and try again
            from coalescenceml.integrations.registry import integration_registry

            integration_registry.activate_integrations()

            try:
                return available_flavors[component_flavor]
            except KeyError:
                raise KeyError(
                    f"No stack component class found for type {component_type} "
                    f"and flavor {component_flavor}. Registered flavors for "
                    f"this type: {set(available_flavors)}. If your stack "
                    f"component class is part of a CoalescenceML integration, make "
                    f"sure the corresponding integration is installed by "
                    f"running `coml integration install INTEGRATION_NAME`."
                ) from None


C = TypeVar("C", bound=StackComponent)


def register_stack_component_class(cls: Type[C]) -> Type[C]:
    """Registers the stack component class and returns it unmodified."""
    StackComponentClassRegistry.register_class(component_class=cls)
    return cls


StackComponentClassRegistry.register_class(LocalOrchestrator)
StackComponentClassRegistry.register_class(SQLiteMetadataStore)
StackComponentClassRegistry.register_class(MySQLMetadataStore)
StackComponentClassRegistry.register_class(LocalArtifactStore)
StackComponentClassRegistry.register_class(BaseContainerRegistry)



from __future__ import annotations

from typing import TYPE_CHECKING, List


if TYPE_CHECKING:
    from coalescenceml.step import BaseStepConfig


class StepContextError(Exception):
    """Raises exception when interacting with a StepContext
    in an unsupported way."""


class StepInterfaceError(Exception):
    """Raises exception when interacting with the Step interface
    in an unsupported way."""


class MissingStepParameterError(Exception):
    """Raises exceptions when a step parameter is missing when running a
    pipeline."""

    def __init__(
        self,
        step_name: str,
        missing_parameters: List[str],
        config_class: BaseStepConfig,
    ):
        """
        Initializes a MissingStepParameterError object.
        Args:
            step_name: Name of the step for which one or more parameters
                       are missing.
            missing_parameters: Names of all parameters which are missing.
            config_class: Class of the configuration object for which
                          the parameters are missing.
        """
        message = f"""
            Missing parameters {missing_parameters} for '{step_name}' step.
            There are two ways to solve this issue:
            (1) Specify a default value in the configuration class
            `{config_class.__name__}`
            (2) Specify the parameters in code when creating the pipeline:
            `my_pipeline({step_name}(config={config_class.__name__}(...))`
            """
        super().__init__(message)



"""
## Container Registry

A container registry is a store for (Docker) containers. A CoalescenceML workflow
involving a container registry would automatically containerize your code to
be transported across stacks running remotely
"""
from coalescenceml.container_registry.base_container_registry import (
    BaseContainerRegistry,
)


__all__ = ["BaseContainerRegistry"]



import json
from pathlib import Path
from typing import Any, Dict

from coalescenceml.io import fileio, utils


def write_json(file_path: str, contents: Dict[str, Any]) -> None:
    """Write contents as JSON to file_path.

    Args:
        file_path: Path to JSON file.
        contents: Contents of json as dict.

    Raises:
        FileNotFoundError: if parent directory of  file_path does not exist
    """
    if not utils.is_remote(file_path):
        # If it is a local path
        directory = str(Path(file_path).parent)
        if not fileio.isdir(directory):
            # If it doesn't exist, then raise exception
            raise FileNotFoundError(f"Directory '{directory}' does not exist")

    utils.write_file_contents_as_string(file_path, json.dumps(contents))


def read_json(file_path: str) -> Dict[str, Any]:
    """Read JSON at file path and return contents.

    Args:
        file_path: path to JSON file.

    Returns:
        Contents of the file as a dict

    Raises:
        FileNotFoundError: if file path does not exist
    """
    if fileio.exists(file_path):
        contents = utils.read_file_contents_as_string(file_path)
        return json.loads(contents)
    else:
        raise FileNotFoundError(f"File '{file_path}' does not exist")



import os
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Iterable,
    List,
    Optional,
    Set,
    Tuple
)

from adlfs import AzureBlobFileSystem
from azure.identity.aio import DefaultAzureCredential

from coalescenceml.artifact_store.base_artifact_store import BaseArtifactStore, PathType
from coalescenceml.integrations.constants import AZURE
from coalescenceml.stack.stack_component_class_registry import (
    register_stack_component_class
)


@register_stack_component_class
class AzureArtifactStore(BaseArtifactStore):
    """ Artifact Store for Azure Datalake artifacts""" 

    account_name: str
    _abfs: Optional[AzureBlobFileSystem] = None

    FLAVOR: ClassVar[str] = AZURE
    SUPPORTED_SCHEMES: ClassVar[Set[str]] = {"az://", "abfs://"}
    
    @property 
    def filesystem(self):
        if not self._abfs:
            creds = DefaultAzureCredential()
            self._abfs = AzureBlobFileSystem(account_name=self.account_name, credential = creds, anon = False)
        return self._abfs

    def open(self, name: PathType, mode: str = "rb") -> Any:
        """ Open a file at""" 
        return self.filesystem.open(name, mode=mode)
    
    def copyfile(self, src: PathType, dst: PathType, overwrite: bool = False) -> None:
        if not overwrite and self.exists(dst):
            raise FileExistsError(
                f"Destination file {str(dst)} already exists and argument "
                f"`overwrite` is false."
            )
        return self.filesystem.cp_file(src, dst) 

    def exists(self, path: PathType) -> bool:
        return self.filesystem.exists(path)

    def glob(self, pattern: PathType) -> List[PathType]:
        return self.filesystem.glob(pattern) 
    
    def isdir(self, path: PathType) -> bool:
        return self.filesystem.isdir(path) 
     
    def listdir(self, path: PathType) -> List[PathType]:
        return self.filesystem.ls(path)

    def makedirs(self, path: PathType) -> None:
        return self.filesystem.mkdir(path)
    
    def mkdir(self, path: PathType) -> None:
        return self.filesystem.mkdir(path, create_parents = False)

    def remove(self, path: PathType) -> None:
        return self.filesystem.rm_file(path)
    
    def rename(self, src: PathType, dst: PathType, overwrite: bool = False) -> None:
        if self.exists(dst) and not overwrite:
            raise FileExistsError(
                f"Destination file {str(dst)} already exists and argument "
                f"`overwrite` is false."
            )
        else:
            self.copyfile(src, dst, overwrite=overwrite) 
            self.remove(src)
            return


    def rmtree(self, path: PathType) -> None:
        return self.filesystem.rm(path)

    def stat(self, path: PathType) -> Any:
        return self.filesystem.info(path)

    def walk(self, top: PathType, topdown: bool = True, onerror: Optional[Callable[..., None]] = None) -> Iterable[Tuple[PathType, List[PathType], List[PathType]]]:
        it = self.filesystem.walk(top)
        if topdown:
            it = list(it)
        else: 
            it = reversed(list(it))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type

from coalescenceml.integrations.exceptions import IntegrationError
from coalescenceml.logger import get_logger


if TYPE_CHECKING:
    from coalescenceml.integrations.integration import Integration

logger = get_logger(__name__)


class IntegrationRegistry(object):
    """Registry to keep track of CoalescenceML Integrations"""

    def __init__(self) -> None:
        """Initializing the integration registry"""
        self._integrations: Dict[str, Type["Integration"]] = {}

    @property
    def integrations(self) -> Dict[str, Type["Integration"]]:
        """Method to get integrations dictionary.
        Returns:
            A dict of integration key to type of `Integration`.
        """
        return self._integrations

    @integrations.setter
    def integrations(self, i: Any) -> None:
        """Setter method for the integrations property"""
        raise IntegrationError(
            "Please do not manually change the integrations within the "
            "registry. If you would like to register a new integration "
            "manually, please use "
            "`integration_registry.register_integration()`."
        )

    def register_integration(
        self, key: str, type_: Type["Integration"]
    ) -> None:
        """Method to register an integration with a given name"""
        self._integrations[key] = type_

    def activate_integrations(self) -> None:
        """Method to activate the integrations with are registered in the
        registry"""
        for name, integration in self._integrations.items():
            if integration.check_installation():
                integration.activate()
                logger.debug(f"Integration `{name}` is activated.")
            else:
                logger.debug(f"Integration `{name}` could not be activated.")

    @property
    def list_integration_names(self) -> List[str]:
        """Get a list of all possible integrations"""
        return [name for name in self._integrations]

    def select_integration_requirements(
        self, integration_name: Optional[str] = None
    ) -> List[str]:
        """Select the requirements for a given integration
        or all integrations"""
        if integration_name:
            if integration_name in self.list_integration_names:
                return self._integrations[integration_name].REQUIREMENTS
            else:
                raise KeyError(
                    f"Version {integration_name} does not exist. "
                    f"Currently the following integrations are implemented. "
                    f"{self.list_integration_names}"
                )
        else:
            return [
                requirement
                for name in self.list_integration_names
                for requirement in self._integrations[name].REQUIREMENTS
            ]

    def is_installed(self, integration_name: Optional[str] = None) -> bool:
        """Checks if all requirements for an integration are installed"""
        if integration_name in self.list_integration_names:
            return self._integrations[integration_name].check_installation()
        elif not integration_name:
            all_installed = [
                self._integrations[item].check_installation()
                for item in self.list_integration_names
            ]
            return all(all_installed)
        else:
            raise KeyError(
                f"Integration '{integration_name}' not found. "
                f"Currently the following integrations are available: "
                f"{self.list_integration_names}"
            )


integration_registry = IntegrationRegistry()



import os
from typing import Any, Type

from coalescenceml.artifacts import DataAnalysisArtifact, DataArtifact
from coalescenceml.logger import get_logger
from coalescenceml.producers.base_producer import BaseProducer
from coalescenceml.producers.producer_registry import register_producer_class
from coalescenceml.utils import json_utils


logger = get_logger(__name__)
DEFAULT_FILENAME = "data.json"


@register_producer_class
class JSONProducer(BaseProducer):
    """Read/Write JSON files."""

    # since these are the 'correct' way to annotate these types.

    ARTIFACT_TYPES = (
        DataArtifact,
        DataAnalysisArtifact,
    )
    TYPES = (
        int,
        str,
        bytes,
        dict,
        float,
        list,
        tuple,
        bool,
    )

    def handle_input(self, data_type: Type[Any]) -> Any:
        """Reads basic primitive types from json."""
        super().handle_input(data_type)
        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)
        contents = json_utils.read_json(filepath)
        if type(contents) != data_type:
            # TODO: Raise error or try to coerce
            logger.debug(
                f"Contents {contents} was type {type(contents)} but expected "
                f"{data_type}"
            )
        return contents

    def handle_return(self, data: Any) -> None:
        """Handles basic built-in types and stores them as json"""
        super().handle_return(data)
        filepath = os.path.join(self.artifact.uri, DEFAULT_FILENAME)
        json_utils.write_json(filepath, data)



from coalescenceml.artifacts.base_artifact import BaseArtifact


class DataAnalysisArtifact(BaseArtifact):
    """Class for Coalescence Data Analysis artifacts.

    This is a base setup for all things created by things such as
    data profiling, data drift, model drift, etc.
    """

    TYPE_NAME = "DataAnalysisArtifact"
